"""
Simplified Interpolating.

Uses scipy.interpolate.
"""

import warnings

import numpy as np
import uncertainties as unc
import uncertainties.unumpy as unp
from scipy import interpolate as interp
from scipy.interpolate import bisplev, bisplrep

# from smpl import plot as splot
from smpl import data, doc
from smpl.doc import append_doc, append_str

unv = unp.nominal_values
usd = unp.std_devs


# mimick removed interp2d from scipy
# https://scipy.github.io/devdocs/tutorial/interpolate/interp_transition_guide.html
def interp_interp2d(xxr, yyr, zzr, kind="linear"):
    kx = 1
    ky = 1
    if kind == "linear":
        kx = 1
        ky = 1
    elif kind == "cubic":
        kx = 3
        ky = 3
    else:
        raise ValueError("kind must be 'linear' or 'cubic'")
    tck = bisplrep(xxr, yyr, zzr, kx=kx, ky=ky)
    return lambda x, y: bisplev(x, y, tck).T


def identity(x):
    return x


default = {
    "interpolator": [
        "cubic",
        "Use 'cubic' (spline) or 'linear' (cf. <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>).",
    ],
    "order": [3, "Spline order."],
    "pre": [identity, "pre interpolation"],
    "post": [identity, "post interpolation"],
    "interpolate_upper_uncertainty": [
        True,
        "Interpolate only using upper uncertainty.",
    ],
    "interpolate_lower_uncertainty": [
        True,
        "Interpolate only using lower uncertainty.",
    ],
}

# @doc.insert_str("\tDefault kwargs\n\n\t")


@append_doc(data.data_kwargs)
@append_str("\t")
@append_str(doc.array_table(default, init=False))
@append_str(
    doc.array_table({"interpolate_kwargs": ["default", "description"]}, bottom=False)
)
def interpolate_kwargs(kwargs):
    """Set default interpolate_kwargs if not set."""
    kwargs = data.data_kwargs(kwargs)
    for k, v in default.items():
        if k not in kwargs:
            kwargs[k] = v[0]
    return kwargs


def interpolate_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).

    Parameters
    ----------
    **kwargs : optional
        see :func:`interpolate_kwargs`.
    """
    kwargs = interpolate_kwargs(kwargs)
    return data.filtered_data_split(datax, datay, **kwargs)


def interpolate(*data, **kwargs):
    # TODO save spl_upd spl_down values instead of recalculate
    # TODO set eps
    kwargs = interpolate_kwargs(kwargs)
    kwargs["sortbyx"] = False
    _, y, _, dy = interpolate_split(data[0], data[-1], **kwargs)
    ret = None
    if dy is None:
        spl_center = _interpolate(*data[:-1], (y), **kwargs)
        ret = np.vectorize(spl_center, otypes=["float"])
    elif (
        kwargs["interpolate_upper_uncertainty"]
        and kwargs["interpolate_lower_uncertainty"]
    ):
        spl_up = _interpolate(*data[:-1], (y + dy), **kwargs)
        spl_down = _interpolate(*data[:-1], (y - dy), **kwargs)
        ret = np.vectorize(
            lambda *a: unc.ufloat(
                spl_up(*a) / 2 + spl_down(*a) / 2,
                np.abs(spl_up(*a) - spl_down(*a)) / 2,
            ),
            otypes=["object"],
        )  # symmetrized error...
        # return np.vectorize(Bounds(spl_up, spl_down), otypes=["object"])
    elif not kwargs["interpolate_lower_uncertainty"]:
        spl_center = _interpolate(*data[:-1], (y), **kwargs)
        spl_up = _interpolate(*data[:-1], (y + dy), **kwargs)
        ret = np.vectorize(
            lambda *a: unc.ufloat(spl_center(*a), np.abs(spl_up(*a) - spl_center(*a))),
            otypes=["object"],
        )  # symmetrized error...
    elif not kwargs["interpolate_upper_uncertainty"]:
        spl_center = _interpolate(*data[:-1], (y), **kwargs)
        spl_down = _interpolate(*data[:-1], (y - dy), **kwargs)
        ret = np.vectorize(
            lambda *a: unc.ufloat(
                spl_center(*a), np.abs(spl_down(*a) - spl_center(*a))
            ),
            otypes=["object"],
        )  # symmetrized error...
    else:
        raise ValueError(
            "interpolate_upper_uncertainty and interpolate_lower_uncertainty can't be both False"
        )

    if not check(ret, *data):
        warnings.warn("Bad interpolation. Increase Order or symmetrize error.")
    return ret


def check(f, *args):
    return np.all(np.isclose(unv(f(*args[:-1])), unv(args[-1]), rtol=1e-2))


def _interpolate(*data, **kwargs):
    ret = None
    datay = kwargs["pre"](data[-1])
    if len(data) == 2:
        if kwargs["interpolator"] == "exp":
            ret = _interpolate_exp(*data[:-1], datay)
        else:
            ret = interp.interp1d(*data[:-1], datay, kind=kwargs["interpolator"])
    elif len(data) == 3:
        if kwargs["interpolator"] == "bivariatespline":
            ret = interp.SmoothBivariateSpline(
                *data[:-1], datay, kx=kwargs["order"], ky=kwargs["order"]
            )
        elif kwargs["interpolator"] == "linearnd":
            ret = interp.LinearNDInterpolator(list(zip(*data[:-1])), datay)
        else:
            ret = interp_interp2d(*data[:-1], datay, kind=kwargs["interpolator"])
    elif kwargs["interpolator"] == "linear":
        ret = interp.LinearNDInterpolator(list(zip(*data[:-1])), datay)

    # return np.vectorize(Post(ret, kwargs['post']), otypes=["object"])
    return np.vectorize(lambda *a: kwargs["post"](ret(*a)), otypes=["object"])


def _interpolate_exp(x, y, **kwargs):
    ip = interp.interp1d(x, np.log(y), kind="linear")
    return lambda x_new: np.exp(ip(x_new))



import re
from io import StringIO

from smpl import doc

from .read_buffer import ReadBuffer


def grep(pattern, *inps, regex=False, open=True, A=0, B=0):
    """
    Searches for ``pattern`` in ``inp``.

    >>> from smpl import io
    >>> io.write("test.txt","hi\\nho1\\n2\\n3\\n4\\n")
    >>> grep("h","test.txt").read()
    'hi\\nho1\\n'
    >>> grep("h.*\\\\d","test.txt",regex=True).read()
    'ho1\\n'
    """
    r = StringIO()
    for inp in inps:
        with ReadBuffer(inp, open=open) as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                match = False
                for j in range(
                    (i - A if float("inf") != A else 0),
                    (i + B + 1 if float("inf") != B else len(lines)),
                ):
                    if j < 0 or j >= len(lines):
                        continue
                    if pattern in lines[j] or (regex and re.search(pattern, lines[j])):
                        match = True
                if match:
                    r.write(line)
    r.seek(0, 0)
    return r


grepf = doc.deprecated(
    version="1.0.6.1",
    removed_in="2.0.0",
    reason="Use :func:`smpl.io.grep(..., open=True)` instead.",
)(grep)


def between(pattern1, pattern2, *inps, regex=False, open=True):
    """
    Searches for ``pattern1`` and ``pattern2`` in ``inp`` and returns the lines between them.

    >>> from smpl import io
    >>> io.write("test.txt","hi\\nho1\\n2\\n3\\n4\\n")
    >>> between("hi","2","test.txt").read()
    'ho1\\n'
    """
    r = StringIO()
    for inp in inps:
        with ReadBuffer(inp, open=open) as f:
            lines = f.readlines()
            start = -1
            for i, line in enumerate(lines):
                if pattern1 in line or (regex and re.search(pattern1, line)):
                    start = i
                if pattern2 in line or (regex and re.search(pattern2, line)):
                    end = i
                    # only write after a closed block if a new block is found before
                    if start > -1:
                        r.write("".join(lines[start + 1 : end]))
                    start = -1
    r.seek(0, 0)
    return r



"""Simplified internal data parsing and transforming."""
import numpy as np
import uncertainties.unumpy as unp

from smpl import doc, stat, util

default = {
    "frange": [
        None,
        "Limit the fit to given range. First integer is the lowest and second the highest index.",
    ],
    "fselector": [
        None,
        "Function that takes ``x`` and ``y`` as parameters and returns an array mask in order to limit the data points for fitting. Alternatively a mask for selecting elements from datax and datay.",
    ],
    "sortbyx": [
        True,
        "Enable sorting the x and y data so that x is sorted.",
    ],
    "bins": [
        0,
        "Number of bins for histogram",
    ],
    "binunc": [
        stat.poisson_dist,
        "Number of bins for histogram",
    ],
    "xerror": [True, "enable xerrors"],
    "yerror": [True, "enable yerrors"],
}


unv = unp.nominal_values
usd = unp.std_devs


@doc.append_str(doc.array_table(default, init=False))
@doc.append_str(
    doc.array_table({"data_kwargs": ["default", "description"]}, bottom=False)
)
@doc.append_str("\n\n")
def data_kwargs(kwargs):
    """Set default :func:`data_kwargs` if not set."""
    for k, v in default.items():
        if not k in kwargs:
            kwargs[k] = v[0]
    return kwargs


def __data_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).
    Does not apply filters `frange` and `fselector`.

    Parameters
    ----------
    **kwargs : optional
        see :func:`data_kwargs`.
    """
    if kwargs["bins"] > 0:
        N, bins = np.histogram(unv(datax), bins=kwargs["bins"])
        y = kwargs["binunc"](N)
        yerr = usd(y)
        yerr = yerr if np.any(np.abs(yerr) > 0) else None
        return bins[0:-1] - (bins[0] - bins[1]) / 2, unv(y), None, yerr
    if util.has("sortbyx", kwargs) and kwargs["sortbyx"]:
        ind = np.argsort(unv(datax))
    else:
        ind = np.array(range(len(datax)))
    x = unv(datax)[ind]
    y = unv(datay)[ind]
    xerr = usd(datax)[ind]
    yerr = usd(datay)[ind]
    xerr = xerr if np.any(np.abs(xerr) > 0) else None
    yerr = yerr if np.any(np.abs(yerr) > 0) else None
    if util.has("xerror", kwargs) and not kwargs["xerror"]:
        xerr = None
    if util.has("yerror", kwargs) and not kwargs["yerror"]:
        yerr = None
    return x, y, xerr, yerr


def _data_split(datax, datay, **kwargs):
    """
    Applies `fselector` and calls :func:`data_split`
    """
    if util.has("fselector", kwargs):
        sel = kwargs["fselector"]
        if callable(sel):
            return __data_split(
                datax[sel(datax, datay)], datay[sel(datax, datay)], **kwargs
            )
        else:
            return __data_split(datax[sel], datay[sel], **kwargs)
    return __data_split(datax, datay, **kwargs)


def filtered_data_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).
    Applies filters `fselector` and `frange`.

    Returns
    -------
    (x,y,xerr,yerr) : tuple
        four arrays with specified values.

    Parameters
    ----------
    **kwargs : optional
        see :func:`data_kwargs`.
    """
    kwargs = data_kwargs(kwargs)
    x, y, xerr, yerr = _data_split(datax, datay, **kwargs)
    if util.has("frange", kwargs):
        x = x[kwargs["frange"][0] : kwargs["frange"][1]]
        y = y[kwargs["frange"][0] : kwargs["frange"][1]]
        if not yerr is None:
            yerr = yerr[kwargs["frange"][0] : kwargs["frange"][1]]
        if not xerr is None:
            xerr = xerr[kwargs["frange"][0] : kwargs["frange"][1]]

    return x, y, xerr, yerr


def flatmesh(*args):
    """
    Similar to `numpy.meshgrid` but the result will be of one dimension instead of stacked arrays.
    """
    r = np.meshgrid(*args)
    rr = []
    for ri in r:
        rr += [ri.reshape(ri.size)]
    return rr



import matplotlib.pyplot as plt
import numpy as np
import pytest

from smpl import plot


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_data():
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    plot.data(x, y)
    return plt.gcf()


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_yscale():
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    plot.data(x, y, yscale="log")
    return plt.gcf()


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_xscale():
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    plot.data(x, y, xscale="log")
    return plt.gcf()


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_grid():
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    plot.data(x, y, grid=False)
    return plt.gcf()



import re
from io import StringIO

from .read_buffer import ReadBuffer


def sed(pattern, replace, *inps, open=True):
    """
    Replace ``pattern`` in ``inp``.

    >>> from smpl import io
    >>> io.write("test.txt","hi\\n")
    >>> sed("hi","cool","test.txt").read()
    'cool\\n'
    """
    r = StringIO()
    for inp in inps:
        with ReadBuffer(inp, open=open) as f:
            lines = f.readlines()
            for _, line in enumerate(lines):
                r.write(re.sub(pattern, replace, line))
    r.seek(0, 0)
    return r



import matplotlib.pyplot as plt
import numpy as np
import smplr
from matplotlib import colors
from matplotlib.image import NonUniformImage

from smpl import doc, util
from smpl import plot as splot

default = {
    "title": [None, "Plot title"],
    "xlabel": [None, "."],
    "ylabel": [None, "."],
    "zlabel": [None, "."],
    "logz": [True, "Colorbar in logarithmic scale."],
    "style": [
        "pcolormesh",
        "Plot via an image ('image') or scatter ('scatter') or mesh ('pcolormesh').",
    ],
    "interpolation": [
        "nearest",
        "Only 'nearest' or 'bilinear' for nonuniformimage. Check https://matplotlib.org/stable/gallery/images_contours_and_fields/interpolation_methods.html#interpolations-for-imshow",
    ],
    "cmap": [
        "viridis",
        "Good default color map for missing datapoints since it does not include white.",
    ],
    # 'zscale' : [None,"Rescale z values."],
}

# @doc.insert_str("\tDefault kwargs\n\n\t")


@doc.append_str(doc.array_table(default, init=False))
@doc.append_str(
    doc.array_table({"plot2d_kwargs": ["default", "description"]}, bottom=False)
)
def plot2d_kwargs(kwargs):
    """Set default plot2d_kwargs if not set."""
    for k, v in default.items():
        if k not in kwargs:
            kwargs[k] = v[0]
    return kwargs


def plot2d(datax, datay, dataz, **kwargs):
    """
    Creates a 2D-Plot.

    Parameters
    ----------
    **kwargs : optional
        see :func:`plot2d_kwargs`.
    """
    kwargs = plot2d_kwargs(kwargs)
    if "xaxis" in kwargs and ("xlabel" not in kwargs or not kwargs["xlabel"]):
        # warnings.warn("xaxis is deprecated. Use xlabel instead.", DeprecationWarning, 2)
        kwargs["xlabel"] = kwargs["xaxis"]
    if "yaxis" in kwargs and ("ylabel" not in kwargs or not kwargs["ylabel"]):
        # warnings.warn("yaxis is deprecated. Use ylabel instead.", DeprecationWarning, 2)
        kwargs["ylabel"] = kwargs["yaxis"]
    if "zaxis" in kwargs and ("zlabel" not in kwargs or not kwargs["zlabel"]):
        # warnings.warn("zaxis is deprecated. Use zlabel instead.", DeprecationWarning, 2)
        kwargs["zlabel"] = kwargs["zaxis"]

    if util.has("axes", kwargs) and kwargs["axes"] is not None:
        plt.sca(kwargs["axes"])
    if kwargs["style"] == "pcolormesh":
        pcolormesh_vplot(datax, datay, dataz, **kwargs)
    elif kwargs["style"] == "image":
        map_vplot(datax, datay, dataz, **kwargs)
    elif kwargs["style"] == "scatter":
        scatter_vplot(datax, datay, dataz, **kwargs)


def sort_xyz(x, y, z):
    p1 = x.argsort(kind="stable")
    x = np.copy(x[p1])
    y = np.copy(y[p1])
    z = np.copy(z[p1])
    p2 = y.argsort(kind="stable")
    x = x[p2]
    y = y[p2]
    z = z[p2]
    return x, y, z


def pcolormesh_vplot(
    tvx,
    tvy,
    tvz,
    xlabel=None,
    ylabel=None,
    zlabel=None,
    logz=True,
    zscale=1.0,
    **kwargs,
):
    """
    Advantage over matplotlibs pcolor(mesh) is that does not require a meshgrid. Instead it uses the data points directly in three lists.
    """
    vx = np.copy(tvx)
    vy = np.copy(tvy)
    vz = np.copy(tvz)
    assert vx.shape == vy.shape == vz.shape

    if len(vz.shape) < 2:
        mesh = np.meshgrid(np.unique(vx), np.unique(vy))
        X, Y = mesh
        # set Z to values of vz on the meshgrid
        Z = np.empty(mesh[0].shape)
        Z[:] = np.nan
        for i, _ in enumerate(vx):
            Z[(mesh[0] == vx[i]) & (mesh[1] == vy[i])] = splot.unv(vz[i])
        Z[:] *= zscale
    else:
        X = vx
        Y = vy
        Z = vz * zscale

    plt.pcolormesh(
        X, Y, Z, norm=colors.LogNorm() if logz else None, cmap=kwargs["cmap"]
    )

    # ax.set_xlim(xl, xm)
    # ax.set_ylim(yl, ym)

    cb = plt.colorbar()
    cb.set_label(zlabel)
    smplr.style_plot2d(xlabel=xlabel, ylabel=ylabel, **kwargs)


def map_vplot(
    tvx,
    tvy,
    tvz,
    xlabel=None,
    ylabel=None,
    zlabel=None,
    logz=True,
    sort=True,
    fill_missing=True,
    zscale=1.0,
    **kwargs,
):
    """ """
    vx = np.copy(tvx)
    vy = np.copy(tvy)
    vz = np.copy(tvz)
    if fill_missing:
        # TODO speed up
        for x in vx:
            for y in vy:
                ex = np.any(np.logical_and((vx == x), (vy == y)))
                if not ex:
                    vx = np.append(vx, x)
                    vy = np.append(vy, y)
                    vz = np.append(vz, 0)
    if sort:
        vx, vy, vz = sort_xyz(vx, vy, vz)

    s = 1
    while vy[s] == vy[s - 1]:
        s = s + 1
    if s == 1:
        # print("flipped x y ")
        while vx[s] == vx[s - 1]:
            s = s + 1
        if s == 1:
            print("error too small map")
            return
        # x, y = y, x
        xlabel, ylabel = ylabel, xlabel
        vx, vy = vy, vx

    grid = splot.unv(vz).reshape((int(np.rint(np.size(vx) / s)), s)) * zscale

    _, ax = plt.subplots(nrows=1, ncols=1, constrained_layout=True)
    im = None
    xl = vx.min() + (vx.min() / 2) - vx[vx != vx.min()].min() / 2
    xm = vx.max() + (vx.max() / 2) - vx[vx != vx.max()].max() / 2
    yl = vy.min() + (vy.min() / 2) - vy[vy != vy.min()].min() / 2
    ym = vy.max() + (vy.max() / 2) - vy[vy != vy.max()].max() / 2
    im = NonUniformImage(
        ax,
        origin="lower",
        cmap=kwargs["cmap"],
        interpolation=kwargs["interpolation"],
        extent=(xl, xm, yl, ym),
        norm=colors.LogNorm() if logz else None,
    )

    im.set_data(np.unique(vx), np.unique(vy), grid)
    ax.add_image(im)
    # ax.images.append(im)
    ax.set_xlim(xl, xm)
    ax.set_ylim(yl, ym)

    cb = plt.colorbar(im)
    cb.set_label(zlabel)
    smplr.style_plot2d(xlabel=xlabel, ylabel=ylabel, **kwargs)


def scatter_vplot(
    vx,
    vy,
    vz,
    xlabel=None,
    ylabel=None,
    zlabel=None,
    logz=True,
    sort=True,
    **kwargs,
):
    if sort:
        vx, vy, vz = sort_xyz(vx, vy, vz)

    _, ax = plt.subplots(nrows=1, ncols=1, constrained_layout=True)
    xl = vx.min() + (vx.min() / 2) - vx[vx != vx.min()].min() / 2
    xm = vx.max() + (vx.max() / 2) - vx[vx != vx.max()].max() / 2
    yl = vy.min() + (vy.min() / 2) - vy[vy != vy.min()].min() / 2
    ym = vy.max() + (vy.max() / 2) - vy[vy != vy.max()].max() / 2

    s = plt.scatter(
        np.concatenate((vx, vx, vx)),
        np.concatenate((vy, vy, vy)),
        c=np.concatenate(
            (
                splot.unv(vz) + splot.usd(vz),
                splot.unv(vz) - splot.usd(vz),
                splot.unv(vz),
            )
        ),
        s=np.concatenate(
            (
                [(3 * plt.rcParams["lines.markersize"]) ** 2 for i in range(len(vx))],
                [(2 * plt.rcParams["lines.markersize"]) ** 2 for i in range(len(vx))],
                [(plt.rcParams["lines.markersize"]) ** 2 for i in range(len(vx))],
            )
        ),
        norm=colors.LogNorm() if logz else None,
        cmap=kwargs["cmap"],
    )

    ax.set_xlim(xl, xm)
    ax.set_ylim(yl, ym)

    cb = plt.colorbar(s)
    cb.set_label(zlabel)
    smplr.style_plot2d(xlabel=xlabel, ylabel=ylabel, **kwargs)



#!/usr/bin/env python3
"""Test script for append_chi2 and append_r2 functionality."""

import numpy as np
import matplotlib.pyplot as plt
from smpl import plot, functions

# Generate some test data
np.random.seed(42)
x = np.linspace(0, 10, 50)
y_true = 2 * x + 3  # linear function: y = 2x + 3
y = y_true + np.random.normal(0, 1, len(x))  # add some noise

# Test 1: Basic fit without chi2/r2
print("Test 1: Basic fit without chi2/r2")
plt.figure(figsize=(12, 4))

# Test 2: Fit with chi2 appended
print("Test 2: Fit with chi2 appended")
params2 = plot.fit(x, y, functions.line, label="With Chi2", append_chi2=True, init=True,fselector=lambda x,y : (x<2))
plt.title("With Chi2")
plt.legend()


plt.tight_layout()
plt.savefig("test_chi2_r2_legend.png", dpi=150, bbox_inches='tight')
plt.show()

print("Test completed successfully!")



"""Simplified parallelization."""

import multiprocessing
from multiprocessing import Process, Queue

import numpy as np

from smpl.doc import append_doc


def queued(q, f, *args, **kwargs):
    q.put(f(*args, **kwargs))


def res(a):
    """
    Parallel evaluation of the list generator from :func:`gen`.

    Return parallel executed values.

    Examples
    --------
    >>> def twice(x):
    ...     return x+x
    >>> for r in [calc(twice,i) for i in range(0,5)]:
    ...     print(res(r))
    0
    2
    4
    6
    8
    >>> res([calc(lambda x : x**3, i) for i in range(0,5)])
    [0, 1, 8, 27, 64]

    """
    if isinstance(a, list):
        return [next(k) for k in a]
    return next(a)


def gen(f, *args, **kwargs):
    """Generates parallel execution list generator."""
    q = Queue()
    p = Process(target=queued, args=(q, f, *args), kwargs=kwargs)
    yield p.start()
    yield [q.get(), p.join()][0]


@append_doc(res)
def calc(f, *args, **kwargs):
    g = gen(f, *args, **kwargs)
    next(g)
    return g


def par(f, *args, **kwargs):
    """
    Parallel execution of f on each element of args and kwargs

    Examples
    --------
    >>> par(lambda x : x**2, range(0,5))
    [0, 1, 4, 9, 16]

    """
    return res(
        [
            calc(
                f,
                *[args[k][i] for k in range(len(args))],
                **{k: v[i] for k, v in kwargs.items()},
            )
            for i in range(
                len(args[0]) if len(args) > 0 else len(next(iter(kwargs.values())))
            )
        ]
    )


# TODO rework to same properties as par
def partitioned_parallel(f, arr, n_jobs=None):
    """
    Parallel execution of f on each element of args

    Examples
    --------
    >>> list(map(int, partitioned_parallel(lambda x : x**2, range(0,5))))
    [0, 1, 4, 9, 16]

    """
    n_jobs = n_jobs or multiprocessing.cpu_count()
    sa = np.array_split(np.array(arr), (len(arr) - 1) // n_jobs + 1)
    res = []
    for i, _ in enumerate(sa):
        res += par(f, sa[i])
    return res


parallel = partitioned_parallel



import matplotlib.pyplot as plt
import numpy as np
import pytest
import uncertainties.unumpy as unp

from smpl import fit
from smpl import functions as f
from smpl.fit import Fitter


def _test_fit_linear(fitter, datax, datay):
    fit.fit(
        datax,
        datay,
        fmt=".",
        function=f.linear,
        units=["l", "b"],
        sigmas=1,
        lpos=2,
        residue=True,
        residue_err=False,
        xaxis="t",
        yaxis="s",
        show=False,
        fitter=fitter,
    )
    plt.close()


def _test_fit_exponential(fitter, datax, datay):
    fit.fit(
        datax,
        datay,
        fmt=".",
        function=f.exp,
        units=["l", "b"],
        sigmas=1,
        lpos=2,
        residue=True,
        residue_err=False,
        xaxis="t",
        yaxis="s",
        b=0,
        show=False,
        fitter=fitter,
    )
    plt.close()


def _test_fit(fitter):
    data = np.loadtxt("tests/test_linear.txt")
    datax, datay = data[:, 0], data[:, 1]
    _test_fit_linear(fitter, datax, datay)
    _test_fit_exponential(fitter, datax, datay)

    datax, datay = unp.uarray(data[:, 0], data[:, 2]), data[:, 1]
    _test_fit_linear(fitter, datax, datay)
    _test_fit_exponential(fitter, datax, datay)

    datax, datay = data[:, 0], unp.uarray(data[:, 1], data[:, 3])
    _test_fit_linear(fitter, datax, datay)
    _test_fit_exponential(fitter, datax, datay)

    datax, datay = (
        unp.uarray(data[:, 0], data[:, 2]),
        unp.uarray(data[:, 1], data[:, 3]),
    )
    _test_fit_linear(fitter, datax, datay)
    _test_fit_exponential(fitter, datax, datay)


def _test_fit_algos():
    _test_fit(Fitter.MINUIT_LEASTSQUARES)
    _test_fit(Fitter.SCIPY_CURVEFIT)
    _test_fit(Fitter.SCIPY_ODR)


@pytest.mark.line_profile.with_args(
    fit.fit,
    fit._fit_minuit_leastsquares,
    fit._fit_curvefit,
    fit._fit_odr,
    _test_fit,
    _test_fit_algos,
)
def test_fit_algos():
    _test_fit_algos()



"""Simplified statistics."""

import math
import statistics as stat
from math import floor, log10

import numpy as np
import pandas as pd
import scipy
import uncertainties as unc
import uncertainties.unumpy as unp
from numpy import arange, array, hstack, newaxis, prod
from scipy import linalg
from scipy.fft import fft as sfft
from scipy.fft import fftfreq, fftshift

from smpl import doc

unv = unp.nominal_values
usd = unp.std_devs


# Copied from scipy.stats._finite_differences
def central_diff_weights(Np, ndiv=1):
    """
    Return weights for an Np-point central derivative.

    Assumes equally-spaced function points.

    If weights are in the vector w, then
    derivative is w[0] * f(x-ho*dx) + ... + w[-1] * f(x+h0*dx)

    Parameters
    ----------
    Np : int
        Number of points for the central derivative.
    ndiv : int, optional
        Number of divisions. Default is 1.

    Returns
    -------
    w : ndarray
        Weights for an Np-point central derivative. Its size is `Np`.

    Notes
    -----
    Can be inaccurate for a large number of points.

    Examples
    --------
    We can calculate a derivative value of a function.

    >>> def f(x):
    ...     return 2 * x**2 + 3
    >>> x = 3.0 # derivative point
    >>> h = 0.1 # differential step
    >>> Np = 3 # point number for central derivative
    >>> weights = central_diff_weights(Np) # weights for first derivative
    >>> vals = [f(x + (i - Np/2) * h) for i in range(Np)]
    >>> float(sum(w * v for (w, v) in zip(weights, vals))/h)
    11.79999999999998

    This value is close to the analytical solution:
    f'(x) = 4x, so f'(3) = 12

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Finite_difference

    """
    if Np < ndiv + 1:
        raise ValueError("Number of points must be at least the derivative order + 1.")
    if Np % 2 == 0:
        raise ValueError("The number of points must be odd.")

    ho = Np >> 1
    x = arange(-ho, ho + 1.0)
    x = x[:, newaxis]
    X = x**0.0
    for k in range(1, Np):
        X = hstack([X, x**k])
    w = prod(arange(1, ndiv + 1), axis=0) * linalg.inv(X)[ndiv]
    return w


# Copied from scipy.stats._finite_differences
def derivative(func, x0, dx=1.0, n=1, args=(), order=3):
    """
    Find the nth derivative of a function at a point.

    Given a function, use a central difference formula with spacing `dx` to
    compute the nth derivative at `x0`.

    Parameters
    ----------
    func : function
        Input function.
    x0 : float
        The point at which the nth derivative is found.
    dx : float, optional
        Spacing.
    n : int, optional
        Order of the derivative. Default is 1.
    args : tuple, optional
        Arguments
    order : int, optional
        Number of points to use, must be odd.

    Notes
    -----
    Decreasing the step size too small can result in round-off error.

    Examples
    --------
    >>> def f(x):
    ...     return x**3 + x**2
    >>> float(derivative(f, 1.0, dx=1e-6))
    4.9999999999...

    """
    if order < n + 1:
        raise ValueError(
            "'order' (the number of points used to compute the derivative), "
            "must be at least the derivative order 'n' + 1."
        )
    if order % 2 == 0:
        raise ValueError(
            "'order' (the number of points used to compute the derivative) must be odd."
        )
    # pre-computed for n=1 and 2 and low-order for speed.
    if n == 1:
        if order == 3:
            weights = array([-1, 0, 1]) / 2.0
        elif order == 5:
            weights = array([1, -8, 0, 8, -1]) / 12.0
        elif order == 7:
            weights = array([-1, 9, -45, 0, 45, -9, 1]) / 60.0
        elif order == 9:
            weights = array([3, -32, 168, -672, 0, 672, -168, 32, -3]) / 840.0
        else:
            weights = central_diff_weights(order, 1)
    elif n == 2:
        if order == 3:
            weights = array([1, -2.0, 1])
        elif order == 5:
            weights = array([-1, 16, -30, 16, -1]) / 12.0
        elif order == 7:
            weights = array([2, -27, 270, -490, 270, -27, 2]) / 180.0
        elif order == 9:
            weights = (
                array([-9, 128, -1008, 8064, -14350, 8064, -1008, 128, -9]) / 5040.0
            )
        else:
            weights = central_diff_weights(order, 2)
    else:
        weights = central_diff_weights(order, n)
    val = 0.0
    ho = order >> 1
    for k in range(order):
        val += weights[k] * func(x0 + (k - ho) * dx, *args)
    return val / prod((dx,) * n, axis=0)


def round_sig(x, sig=2):
    """
    Round to ``sig`` significant digits.

    Parameters
    ----------
    x : float
        Value to round.
    sig : int
        Number of significant digits.

    Returns
    -------
    float
        Rounded value.

    Examples
    --------
    >>> round_sig(1.23456789, sig=2)
    1.2
    >>> round_sig(1.23456789, sig=4)
    1.235
    """
    return round(x, sig - int(floor(log10(abs(x)))) - 1)


def R2(y, f):
    """
    R2 - Coefficient of determination

    In the best case, the modeled values exactly match the observed values, which results in R2 = 1.
    A baseline model, which always predicts the mean of y, will have R2 = 0.
    Models that have worse predictions than this baseline will have a negative R2.

    References
    ----------

    https://en.wikipedia.org/wiki/Coefficient_of_determination
    """
    r = y - f
    mean = np.sum(r) / len(r)
    SSres = np.sum((r) ** 2)
    SStot = np.sum((r - mean) ** 2)
    Rsq = 1 - SSres / SStot
    return Rsq


r2 = R2


def Chi2(y, f, sigmas=None):
    """
    Chi2 - Goodness of Fit

    In general, if Chi-squared/Nd is of order 1.0, then the fit is reasonably good.
    Coversely,  if Chi-squared/Nd >> 1.0, then the fit is a poor one.

    References
    ----------

    https://www.phys.hawaii.edu/~varner/PHYS305-Spr12/DataFitting.html
    """
    r = y - f
    if sigmas is not None:
        chisq = np.sum((r / sigmas) ** 2)
    else:
        chisq = np.sum((r) ** 2)

    return chisq


chi2 = Chi2


def average_deviation(y, f):
    r = np.abs((y - f) / f)
    return mean(r)


def unv_lambda(f):
    """Returns a function which applies :func:`unv` on the result of ``f``."""
    return lambda *a: unv(f(*a))


def poisson_dist(N):
    """
    Return ``N`` with added poissonian uncertainties.

    Parameters
    ----------
    N : float or array_like of floats
        Number of events.

    Returns
    -------
    uncertainties.unumpy.uarray
        Number of events with uncertainties.

    Examples
    --------
    >>> poisson_dist(100)
    array(100.0+/-10.0, dtype=object)
    """
    return unp.uarray(N, np.sqrt(N))


def no_dist(N):
    """Return ``N`` with no uncertainties."""
    return unp.uarray(N, 0)


def normalize(ydata):
    """
    Return normalized ``ydata``.

    Parameters
    ----------
    ydata : array_like
        Data to be normalized.

    Returns
    -------
    array_like
        Normalized data.

    Examples
    --------
    >>> ydata = np.array([1, 2, 3, 4, 5])
    >>> normalize(ydata)
    array([0.  , 0.25, 0.5 , 0.75, 1.  ])
    """
    return (ydata - np.amin(ydata)) / (np.amax(ydata) - np.amin(ydata))


def novar_mean(n):
    """Return mean of ``n`` with only the uncertainties of ``n`` and no variance."""
    return np.sum(n) / len(n)


def mean(n):
    """
    Return mean of ``n`` with combined error of variance and unvertainties of ``n``.

    Parameters
    ----------
    n : array_like
        Data to be averaged.

    Returns
    -------
    uncertainties.unumpy.uarray
        Mean of ``n``.

    Examples
    --------
    >>> n = np.array([1, 2, 3, 4, 5])
    >>> mean(n)
    3.0+/-1.5811388300841898
    """
    # find the mean value and add uncertainties
    if isinstance(n, pd.core.series.Series):
        n = n.to_numpy()
    k = np.mean(n)
    err = stat.variance(unv(n))
    return unc.ufloat(unv(k), math.sqrt(usd(k) ** 2 + err))


def noisy(x, mean=1, std=0.1):
    """
    Add gaussian noise to ``x``.

    Parameters
    ----------
    x : array_like
        Data to be smeared.
    mean : float
        Mean of gaussian noise.
    std : float
        Standard deviation of gaussian noise.

    Returns
    -------
    array_like
        Smeared data.

    Examples
    --------
    >>> x = np.array([1, 2, 3, 4, 5])
    >>> noisy(x,std=0)
    array([1., 2., 3., 4., 5.])
    """
    return x * np.random.normal(mean, std, len(x))


def normal(x, mean=0, std=1):
    return np.random.normal(mean, std, len(x))


@doc.insert_eq()
def fft(y):
    """
    Compute the FFT of ``y``.

    Parameters
    ----------
    y : array_like
        Data to be transformed.

    Returns
    -------
    array_like

    """
    t = y
    sp = fftshift(sfft(np.sin(t)))
    freq = fftshift(fftfreq(t.shape[-1]))
    return freq, sp


def trim_domain(
    f,
    fmin=np.finfo(np.float32).min / 2,
    fmax=np.finfo(np.float32).max / 2,
    steps=10000,
    min_ch=0.0001,
    recursion_limit=10,
):
    """
    Get the domain of the function ``f`` with the ranges removed where the derivative of ``f`` is below ``min_ch``.
    """
    recursion_limit = recursion_limit - 1
    if recursion_limit < 0:
        return fmin, fmax
    test = np.linspace(fmin, fmax, steps)
    try:
        dr = derivative(f, test, dx=1e-06)
    except Exception:
        return 0.0, 0.0
    m1 = np.abs(dr) > min_ch
    bmin = np.argmax(m1)
    m2 = (np.abs(dr) > min_ch)[::-1]
    tbmax = np.argmax(m2)
    xmin = test[bmin]
    xmax = test[::-1][tbmax]
    if bmin == 0 and tbmax == 0 and not m1[0] and not m2[0]:
        # trisect the full domain
        tmin = xmin
        tmax = xmax
        t1a, t1b = trim_domain(
            f,
            tmin + (tmax - tmin) / 3,
            tmax - (tmax - tmin) / 3,
            min_ch=min_ch,
            recursion_limit=recursion_limit,
        )
        if np.isclose(t1a, t1b):
            t2a, t2b = trim_domain(
                f,
                tmin + (tmax - tmin) / 3,
                tmax,
                min_ch=min_ch,
                recursion_limit=recursion_limit,
            )
            if np.isclose(t2a, t2b):
                t3a, t3b = trim_domain(
                    f,
                    tmin,
                    tmax - (tmax - tmin) / 3,
                    min_ch=min_ch,
                    recursion_limit=recursion_limit,
                )
                if np.isclose(t3a, t3b):
                    return 0.0, 0.0
                return t3a, t3b
            return t2a, t2b
        return t1a, t1b
    return xmin, xmax


def get_domain(
    f,
    fmin=np.finfo(np.float32).min / 2,
    fmax=np.finfo(np.float32).max / 2,
    steps=1000,
):
    """
    Return the statistically probed domain of the function ``f``.
    """
    if np.isclose(fmin, fmax, rtol=0.0001, atol=0.00001):
        return 0.0, 0.0

    test = np.linspace(fmin, fmax, steps)

    r = unv(f(test))
    mask = np.isfinite(r)
    tr = test[mask]
    if len(tr) > 0:
        tmin = np.amin(tr)
        tmax = np.amax(tr)
        test_r = np.linspace(tmin, tmax, steps)
        if np.equal(tr.shape, test_r.shape) and np.allclose(test_r, tr):
            return tmin, tmax

    # trisect
    tmin = fmin
    tmax = fmax
    t1a, t1b = get_domain(f, tmin + (tmax - tmin) / 3, tmax - (tmax - tmin) / 3)
    if np.isclose(t1a, t1b):
        t2a, t2b = get_domain(f, tmin + (tmax - tmin) / 3, tmax)
        if np.isclose(t2a, t2b):
            t3a, t3b = get_domain(f, tmin, tmax - (tmax - tmin) / 3)
            if np.isclose(t3a, t3b):
                return 0.0, 0.0
            return t3a, t3b
        return t2a, t2b
    return t1a, t1b


def is_monotone(f, tmin=None, tmax=None, steps=1000):
    """
    Test if function ``f`` is monotone.

    Parameters
    ----------
    f : function
        Function to be tested.
    test : array_like
        Test points.

    Returns
    -------
    bool
        True if function is monotone.

    Examples
    --------
    >>> def f(x):
    ...     return x**2
    >>> is_monotone(f)
    False
    >>> is_monotone(np.exp)
    True
    """
    if tmax is None and tmin is None:
        tmin, tmax = get_domain(f)
    test = np.linspace(tmin, tmax, steps)
    return bool(np.all(f(test[1:]) >= f(test[:-1])))


def get_interesting_domain(f, min_ch=1e-6, maxiter=100):
    """
    Return interesting xmin and xmax of function ``f``.

    Examples
    --------
    >>> def f(x):
    ...     return np.sin(x)
    >>> get_interesting_domain(f)
    (-3.141625000000003, 3.141625000000003)
    """
    omin_x, omax_x = get_domain(f)
    if is_monotone(f, omin_x, omax_x):
        min_x, max_x = trim_domain(f, omin_x, omax_x, min_ch=min_ch)
        # min_x,max_x=omin_x,omax_x
    else:
        tmax_x = scipy.optimize.minimize(
            lambda x: -f(x),
            0.0,
            method="Nelder-Mead",
            bounds=[(omin_x, omax_x)],
            options={"maxiter": maxiter},
        )
        tmin_x = scipy.optimize.minimize(
            f,
            0.0,
            method="Nelder-Mead",
            bounds=[(omin_x, omax_x)],
            options={"maxiter": maxiter},
        )
        if tmax_x.success:
            tmax_x = tmax_x.x[0]
        else:
            tmax_x = 0.0
        if tmin_x.success:
            tmin_x = tmin_x.x[0]
        else:
            tmin_x = 0.0

        if abs(tmax_x) > np.finfo(np.float32).max / 10:
            tmax_x = 0.0
        if abs(tmin_x) > np.finfo(np.float32).max / 10:
            tmin_x = 0.0
        x_min = min(tmax_x, tmin_x)
        x_max = max(tmax_x, tmin_x)
        min_x = (x_max + x_min) / 2 - (x_max - x_min)
        max_x = (x_max + x_min) / 2 + (x_max - x_min)
        if np.isclose(min_x, max_x):
            min_x, max_x = trim_domain(f, omin_x, omax_x, min_ch=min_ch)

    return float(min_x), float(max_x)



from .doc import *

## Functions below are not in the smpl.doc package since they are very specific but fit in doc perfectly.


def append_plot(*args, xmin=-5, xmax=5):
    """Append a plot to a function."""
    # return _append("\n\n\t.. plot::\n\t\t:include-source:\n\n\t\t>>> from " + target.__module__ + " import " +target.__name__ + "\n\t\t>>> from smpl import plot\n\t\t>>> plot.function("+ target.__name__ + "," + ','.join([str(a) for a in args]) + ",xmin="+str(xmin) + ",xmax=" + str(xmax)+")")

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ += (
            "\n\n\t.. plot::\n\t\t:include-source:\n\n\t\t>>> from "
            + target.__module__
            + " import "
            + target.__name__
            + "\n\t\t>>> from smpl import plot\n\t\t>>> plot.function("
            + target.__name__
            + ("," + ",".join([str(a) for a in args]) if len(args) > 0 else "")
            + ",xmin="
            + str(xmin)
            + ",xmax="
            + str(xmax)
            + ")"
        )
        # print(target.__doc__)
        return target

    return wrapper


def insert_eq():
    """Inserts the function and its parameters and an equal sign."""

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        safe = target.__doc__
        target.__doc__ = target.__name__ + "("
        for v in target.__code__.co_varnames:
            target.__doc__ += v + ","
        target.__doc__ = target.__doc__[:-1]
        target.__doc__ += ") = " + safe
        return target

    return wrapper


def insert_latex_eq():
    """Inserts latexed code of a oneline function with parameters."""
    return lambda f: insert_eq()(insert_latex()(f))


def insert_latex():
    """Inserts latexed code of a oneline function."""
    from smpl import wrap

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = wrap.get_latex(target)
        return target

    return wrapper



import inspect
import warnings
from tokenize import TokenError

import numpy as np
import sympy
import uncertainties.unumpy as unp
from sympy.parsing.sympy_parser import (
    implicit_multiplication_application,
    standard_transformations,
)
from sympy.printing.pycode import pycode


def get_varnames(expr, xvar):
    """
    Returns a list of variables used in the ``str`` math-expression via sympy and puts ``xvar`` to the front.

    Examples
    --------
    >>> get_varnames("a**x*b+c","x")
    ['x', 'a', 'b', 'c']

    """
    if isinstance(expr, str):
        return str_get_varnames(expr, xvar)
    else:
        return fnc_get_varnames(expr, xvar)


def get_latex(function):
    """
    Return a latex string for passed function.


    Parameters
    ----------
    function : function_like
        function as str lambda or (oneline) function

    Examples
    --------
    >>> get_latex(lambda a,b,c,x : (a+b+c)*x,)
    '$x \\\\left(a + b + c\\\\right)$'
    >>> get_latex("(a+b+c)*x")
    '$x \\\\left(a + b + c\\\\right)$'
    >>> def fun(a,b,x,c):
    ...     return (a+b+c)*x
    >>> get_latex(fun)
    '$x \\\\left(a + b + c\\\\right)$'

    """
    if isinstance(function, str):
        l = "$" + sympy.latex(str_get_expr(function)) + "$"
    else:
        l = function.__name__
    if l == "<lambda>":
        try:
            try:
                cc, li = inspect.findsource(function)
                f = (
                    "".join(cc[li:])
                    .split("lambda")[1]
                    .split(":")[1]
                    .split(",")[0]
                    .replace("\n", "")
                )
                l = "$" + sympy.latex(str_get_expr(f)) + "$"
            except TokenError:
                raise Exception(
                    "Make sure there is a ',' behind the lambda expression and no commas are in the lambda expression and no newlines"
                )
        except OSError:
            l = "$\\lambda$(" + ",".join(function.__code__.co_varnames) + ")"
    elif not isinstance(function, str):
        if function.__doc__ is not None:
            l = function.__doc__.split("\n")[0]
        else:
            try:
                cc, li = inspect.findsource(function)
                s = "".join(cc[li:]).split("return")[1].split("\n")[0]
                l = "$" + sympy.latex(str_get_expr(s)) + "$"
            except OSError:
                l = function.__name__

    return l


def get_lambda_argd(expr, xvar, *args):
    function = get_lambda(expr, xvar)
    return lambda x: function(x, *args)


def get_lambda(expr, xvar):
    """
    Returns a lambda of given ``str``/``function``/``lambda`` expression with ``__doc__`` set to the latex expression. ``xvar`` is moved to the front.

    Examples
    --------
    >>> l = get_lambda(lambda a,b,c,x : (a+b+c)*x,'x')
    >>> l(4,1,1,1)
    12
    >>> l = get_lambda("(a+b+c)*x",'x')
    >>> l(4,1,1,1)
    12
    >>> def fun(a,b,x,c):
    ...     return (a+b+c)*x
    >>> l = get_lambda(fun,'x')
    >>> l(4,1,1,1)
    12

    """
    if isinstance(expr, str):
        return str_get_lambda(expr, xvar)
    else:
        return fnc_get_lambda(expr, xvar)


def fnc_get_lambda(expr, xvar):
    __l__ = eval(
        "lambda "
        + ",".join(get_varnames(expr, xvar))
        + ": expr("
        + ",".join(expr.__code__.co_varnames)
        + ")",
        {"expr": expr},
    )
    __l__.__doc__ = expr.__doc__
    return __l__


def str_get_lambda(expr, xvar):
    parsed_expr = str_get_expr(expr)
    pc = pycode(parsed_expr)
    for s in unp.__all__:
        pc = pc.replace("math." + s, "unp." + s)
    # old direct way. Doesn't use unp.
    # __l__ = sympy.lambdify(get_varnames(expr),parsed_expr)
    __l__ = eval("lambda " + ",".join(get_varnames(expr, xvar)) + ": " + pc)
    # exec("global __l__; __l__ = lambda " + ','.join(get_varnames(expr)) + ": "+ pc)
    return __l__


def str_get_varnames(expr, xvar):

    parsed_expr = str_get_expr(expr)

    new_locals = [sym.name for sym in parsed_expr.atoms(sympy.Symbol)]
    new_locals = sorted(new_locals)
    if xvar is None:
        return np.roll(new_locals, 1)
    else:
        varss = new_locals
        assert xvar in varss
        varss.remove(xvar)
        return [xvar, *varss]


def fnc_get_varnames(func, xvar):
    if xvar is None:
        return func.__code__.co_varnames
    else:
        varss = [s for s in func.__code__.co_varnames]
        assert xvar in varss
        varss.remove(xvar)
        return [xvar, *varss]


def str_get_expr(expr):
    """
    Convert a pythonic string expression ot a sympy expression.

    Only works with np or unp naming.
    """
    expr = (
        expr.replace("math.abs(", "Abs(")
        .replace("np.abs(", "Abs(")
        .replace("unp.abs(", "Abs(")
    )
    expr = expr.replace("math.", "").replace("unp.", "").replace("np.", "")
    try:
        parsed_expr = sympy.parsing.sympy_parser.parse_expr(
            expr,
            local_dict=None,
            transformations=(
                standard_transformations + (implicit_multiplication_application,)
            ),
            evaluate=False,
        )
    except Exception as e:
        warnings.warn("Illegal variable/function name (try uncap. letters) " + expr)
        raise e
    return parsed_expr



"""Simplified Animations."""

import io
import itertools
import os
import uuid

import ipywidgets as widgets
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation
from matplotlib.backends.backend_agg import FigureCanvasAgg
from PIL import Image

from smpl import doc

frames = []


@doc.deprecated("1.0.5", "Will be removed from main smpl")
def dict_product(dicts):
    """
    >>> d = {"number": [1,2], "color": ['a','b'] }
    >>> list(dict_product(d))
    [{'number': 1, 'color': 'a'}, {'number': 1, 'color': 'b'}, {'number': 2, 'color': 'a'}, {'number': 2, 'color': 'b'}]
    """
    return (dict(zip(dicts, x)) for x in itertools.product(*dicts.values()))


def list_product(lists):
    """
    >>> l = [[1,2],[3,4]]
    >>> list(list_product(l))
    [(1, 3), (1, 4), (2, 3), (2, 4)]
    """
    return itertools.product(*lists)


# TODO mirror ipywidgets.interactive options
def interactive(
    func, *args, prerender=True, auto_png=True, rec=0, isls=None, plays=None, **kwargs
):
    if plays is None:
        plays = []
    if isls is None:
        isls = []
    if rec == 0:
        isls = []
        for arg in [*args, *kwargs.values()]:
            r = np.arange(arg.min, arg.max + arg.step, arg.step)
            if isinstance(arg, widgets.Play):
                isl = widgets.IntSlider()
                widgets.jslink((arg, "value"), (isl, "value"))
                plays += [arg]
            else:
                isl = widgets.SelectionSlider(
                    options=r,
                    value=arg.value,
                    continuous_update=True,
                    description=arg.description,  # TODO copy more
                )
                plays += [None]
            isls += [isl]
    if not prerender:
        widgets.interactive(func, *args, **kwargs)
    else:
        key = None
        arg = None
        if len(kwargs) == 0:
            arg = args[0]
        if len(args) == 0:
            key, arg = list(kwargs.items())[0]
        outs = []
        r = np.arange(arg.min, arg.max + arg.step, arg.step)
        for a in r:
            tout = widgets.Output(layout={"border": "0px solid black"})
            with tout:
                if (len(args) == 1 and len(kwargs) == 0) or (
                    len(args) == 0 and len(kwargs) == 1
                ):
                    if key is None:
                        func(a)
                    else:
                        func(**{key: a})
                    if auto_png:
                        output = io.BytesIO()
                        plt.savefig(output, format="png")
                        plt.close()
                        # plt.clf()

                        tout = widgets.Image(
                            value=output.getvalue(),
                            format="png",
                        )
                elif len(args) != 0:
                    tout = interactive(
                        lambda *ar, **kw: func(a, *ar, **kw),
                        *args[1:],
                        prerender=prerender,
                        auto_png=auto_png,
                        rec=rec + 1,
                        plays=plays,
                        isls=isls,
                        **kwargs,
                    )
                else:
                    dkw = dict(kwargs).pop(key)
                    tout = interactive(
                        lambda *ar, **kw: func(*ar, **{key: a}, **kw),
                        prerender=prerender,
                        auto_png=auto_png,
                        rec=rec + 1,
                        plays=plays,
                        isls=isls,
                        **dkw,
                    )
            outs += [tout]

        tab = widgets.Tab(children=outs)
        if hasattr(isls[rec], "index"):
            widgets.jslink((isls[rec], "index"), (tab, "selected_index"))
        elif hasattr(isls[rec], "value"):
            widgets.jslink((isls[rec], "value"), (tab, "selected_index"))
        else:
            print("no link")

        if rec:
            return tab
        out = widgets.Output(layout={"border": "0px solid black"})
        for i, isl in enumerate(isls):
            if plays[i] is not None:
                out.append_display_data(widgets.HBox([plays[i], isl]))
            else:
                out.append_display_data(isl)
        out.append_display_data(tab)
        plt.close()
        return out


# class WigAnimation(widget.Image):


class FigAnimation(animation.FuncAnimation):
    def widget_gif(self):
        # convert to gif through save as anim.save wants a filename
        uf = str(uuid.uuid4())
        self.save(uf + ".gif")
        with open(uf + ".gif", "rb") as file:
            image = file.read()
        os.remove(uf + ".gif")

        return widgets.Image(
            value=image,
            format="gif",
        )

    def __init__(self, figs=None, frames=None, init=None, update=None, *args, **kwargs):
        # Use the list of figures as the framedata, which will be iterated
        # over by the machinery.
        if update is None:
            f1 = figs[0]
        else:
            update(frames[0])
            f1 = plt.gcf()
            plt.clf()
            plt.close()

        self._figs = figs
        f = plt.figure()
        plt.axis("off")
        plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
        plt.margins(0, 0)

        canvas = FigureCanvasAgg(f1)
        canvas.draw()
        rgba = np.asarray(canvas.buffer_rgba())
        im = Image.fromarray(rgba)
        self.im = plt.imshow(im)

        def _update(frame):
            ff = None
            if update is not None:
                update(frame)
                ff = plt.gcf()
            else:
                ff = figs[frame]
            canvas = FigureCanvasAgg(ff)
            canvas.draw()
            rgba = np.asarray(canvas.buffer_rgba())
            im = Image.fromarray(rgba)
            self.im.set_array(mpl.image.pil_to_array(im))
            return (self.im,)

        animation.FuncAnimation.__init__(
            self,
            f,
            _update,
            frames=len(figs) if update is None else frames,
            init_func=init,
            *args,
            **kwargs,
        )


def frame():
    """
    Saves current Matplotlib figure.
    """
    global frames
    # f = plt.gcf()
    # f.figure = f
    # diold = f.canvas.draw_idle
    # f.set_visible = lambda b: vis(f, b)
    # f.canvas.draw_idle = lambda _=None: io.pr(
    #    diold()) if len(frames) == 0 else None

    frames.append(plt.gcf())
    # plt.savefig("test"+str(len(frames))+".jpg")
    plt.close()


def clear():
    """
    Empties stored frames.
    """
    global frames
    frames = []


def animate(**kwargs):
    """
    Make frames to Animation

    Parameters
    ==========

    They are passed directly to ArtistAnimation.

    Returns
    =======
    ArtistAnimation

    """
    global frames
    ani = FigAnimation(frames, **kwargs)
    # clear()
    return ani



import matplotlib.pyplot as plt
import numpy as np
import pytest

from smpl import plot


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_style_scatter():
    x = np.linspace(0, 10, 10)
    y = np.sin(x)
    z = np.abs(np.cos(x))
    plot.plot2d(x, y, z, style="scatter")
    return plt.gcf()


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_style_pcolormesh():
    x = np.linspace(0, 10, 10)
    y = np.sin(x)
    z = np.abs(np.cos(x))
    plot.plot2d(x, y, z, style="pcolormesh")
    return plt.gcf()


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_style_image():
    x = np.linspace(0, 10, 10)
    y = np.sin(x)
    z = np.abs(np.cos(x))
    plot.plot2d(x, y, z, style="image")
    return plt.gcf()



"""
Simplified Fitting.

Uses scipy.curve_fit (no x errors) or scipy.odr (with x errors).
"""

import enum

import numpy as np
import uncertainties.unumpy as unp
from numpy.linalg import LinAlgError
from tqdm import tqdm

from smpl import data, debug, doc, functions, stat, util, wrap
from smpl.util.util import dict_unique_by_value

unv = unp.nominal_values
usd = unp.std_devs


class Fitter(enum.Enum):
    """
    Different implementations to perform a fit.
    """

    AUTO = 0
    SCIPY_CURVEFIT = 1
    SCIPY_ODR = 2
    MINUIT_LEASTSQUARES = 3


default = {
    "params": [
        None,
        "Initial fit parameters",
    ],
    # 'frange': [None, "Limit the fit to given range. First integer is the lowest and second the highest index.", ],
    # 'fselector': [None, "Function that takes ``x`` and ``y`` as parameters and returns an array mask in order to limit the data points for fitting. Alternatively a mask for selecting elements from datax and datay.", ],
    "fixed_params": [
        True,
        "Enable fixing parameters by choosing the same-named variables from ``kwargs``.",
    ],
    # 'sortbyx': [True, "Enable sorting the x and y data so that x is sorted.", ],
    "maxfev": [
        10000,
        "Maximum function evaluations during fitting.",
    ],
    "epsfcn": [
        0.0001,
        "Suitable step length for jacobian approximation.",
    ],
    "xvar": [
        None,
        "Variable in fit function parameters that corresponds to the x axis. If it is None the last of the alphabetical sorted parameters is used.",
    ],
    # 'bins': [0, "Number of bins for histogram", ],
    # 'binunc': [stat.poisson_dist, "Number of bins for histogram", ],
    "autotqdm": [
        True,
        "Auto fitting display tqdm",
    ],
    # 'xerror': [True, "enable xerrors"],
    # 'yerror': [True, "enable yerrors"],
    "fitter": [Fitter.AUTO, "Choose from :class:`Fitter`s."],
}

# @doc.insert_str("\tDefault kwargs\n\n\t")


@doc.append_doc(data.data_kwargs)
@doc.append_str("\t")
@doc.append_str(doc.array_table(default, init=False))
@doc.append_str(
    doc.array_table({"fit_kwargs": ["default", "description"]}, bottom=False)
)
@doc.append_str("\n\n")
def fit_kwargs(kwargs):
    """Set default :func:`fit_kwargs` if not set."""
    kwargs = data.data_kwargs(kwargs)
    for k, v in default.items():
        if k not in kwargs:
            kwargs[k] = v[0]
    return kwargs


# @append_doc(default_kwargs)
def auto(datax, datay, funcs=None, **kwargs):
    """
    Automatically loop over functions and fit the best one.

    Parameters
    ----------
    funcs : function array
        functions to consider as fit. Default all ``smpl.functions``.
    **kwargs : optional
        see :func:`fit_kwargs`.

    Returns
    -------
    The best fit function and it's parameters and a ``lambda`` where the parameters are already applied to the function.

    """
    kwargs = fit_kwargs(kwargs)
    min_sq = None
    best_f = None
    best_ff = None

    if funcs is None:
        # TODO this contains aliased functions
        funcs = list(dict_unique_by_value(functions.__dict__).values())
        # gives weird fitting errors...
        funcs.remove(functions.fac)
    for f in tqdm(funcs, disable=not kwargs["autotqdm"]):
        if callable(f):
            try:
                ff = fit(datax, datay, f, **kwargs)
                fy = f(datax, *ff)
                sum_sq = (
                    np.sum((fy - datay) ** 2)
                    + np.sum((fy + usd(fy) - datay) ** 2)
                    + np.sum((fy - usd(fy) - datay) ** 2)
                )
            except (ValueError, LinAlgError, OverflowError) as ve:
                debug.msg(ve)
                continue
            if min_sq is None or sum_sq < min_sq:
                min_sq = sum_sq
                best_f = f
                best_ff = ff
    # if not best_f is None:
    #    fit(datax,datay,best_f,**kwargs)
    return best_f, best_ff, lambda x: best_f(x, *best_ff)


def fit(datax, datay, function, **kwargs):
    """
    Returns a fit of ``function`` to ``datax`` and ``datay``.

    Parameters
    ----------
    datax : array_like
        X data either as ``unp.uarray`` or ``np.array`` or ``list``
    datay : array_like
        Y data either as ``unp.uarray`` or ``np.array`` or ``list``
    function : func
        Fit function with parameters: ``x``, ``params``
    **kwargs : optional
        see :func:`fit_kwargs`.

    """
    kwargs = fit_kwargs(kwargs)
    x, y, xerr, yerr = fit_split(datax, datay, **kwargs)

    tmp, params, fixed, Ntot = _wrap_func_and_param(function, **kwargs)

    fitter = kwargs["fitter"]
    if fitter is Fitter.AUTO:
        if xerr is not None:
            fitter = Fitter.SCIPY_ODR
        else:
            fitter = Fitter.SCIPY_CURVEFIT
    if fitter is Fitter.MINUIT_LEASTSQUARES:
        fitt = _fit_minuit_leastsquares(x, y, tmp, params=params, yerr=yerr)
    elif fitter is Fitter.SCIPY_CURVEFIT:
        fitt = _fit_curvefit(x, y, tmp, params=params, yerr=yerr)
    elif fitter is Fitter.SCIPY_ODR:
        fitt = _fit_odr(x, y, tmp, params=params, xerr=xerr, yerr=yerr)
    else:
        raise ValueError(f"Unknown fitter: {fitter}")

    return _unwrap_param(fitt, fixed, Ntot)


def _wrap_func_and_param(function, **kwargs):
    """
    Wraps a function with a lambda function.
    """
    params = None
    if util.has("params", kwargs):
        params = kwargs["params"]
    fixed = {}
    vnames = wrap.get_varnames(function, kwargs["xvar"])
    Ntot = len(vnames) - 1
    if util.has("fixed_params", kwargs) and kwargs["fixed_params"]:
        for i in range(1, len(vnames)):
            if util.has(vnames[i], kwargs):
                fixed[i] = kwargs[vnames[i]]
    # Count parameters for function
    if params is None:
        N = len(vnames)
        params = [1 for i in range(N - 1)]
    tmp_params = []
    for i, pi in enumerate(params):
        if not util.has(i + 1, fixed):
            tmp_params += [pi]
    params = tmp_params
    N = len(params)
    wlambda = wrap.get_lambda(function, kwargs["xvar"])

    def _wrapped_func(x0, *x):
        """
        wrapper for function with fixed parameters applied.
        """
        tmp_x = []
        j = 1
        # print(x)
        for i in range(1, Ntot + 1):
            # print(i," ",j)
            if not util.has(i, fixed):
                tmp_x += [x[j - 1]]
                # print(x[j])
                j = j + 1
            else:
                tmp_x += [fixed[i]]

        # print(Ntot)
        # print(tmp_x)
        return unv(wlambda(x0, *tmp_x))

    return _wrapped_func, params, fixed, Ntot


def _unwrap_param(fitt, fixed, Ntot):
    rfit = []
    j = 0
    for i in range(1, Ntot + 1):
        if not util.has(i, fixed):
            rfit += [fitt[j]]
            j = j + 1
        else:
            rfit += [fixed[i]]
    return rfit


@doc.insert_doc(stat.Chi2)
def Chi2(datax, datay, function, ff, **kwargs):
    kwargs = fit_kwargs(kwargs)
    x, y, _, yerr = fit_split(datax, datay, **kwargs)
    f = function(x, *ff)
    sigmas = (yerr**2 + usd(f) ** 2) ** 0.5
    return stat.Chi2(y, unv(f), sigmas)


def Dof(datax, datay, function, ff, **kwargs):
    """Calculate the degrees of freedom for a fit."""
    kwargs = fit_kwargs(kwargs)
    x, y, _, yerr = fit_split(datax, datay, **kwargs)
    f = function(x, *ff)

    # Obtain the number of fitted parameters
    # by counting the number of non-fixed parameters
    # This assumes that the fit function has parameters
    # in the order: x, param1, param2, ..., paramN
    # and that fixed parameters are provided in kwargs
    _, params, fixed, Ntot = _wrap_func_and_param(function, **kwargs)

    # Number of points fitted
    N_points = len(y)

    # Fixed parameters do not count towards degrees of freedom
    return N_points - Ntot + len(fixed) 

@doc.insert_doc(stat.R2)
def R2(datax, datay, function, ff, **kwargs):
    kwargs = fit_kwargs(kwargs)
    x, y, _, _ = fit_split(datax, datay, **kwargs)
    return stat.R2(y, unv(function(x, *ff)))


def data_split(datax, datay, **kwargs):
    """
    Split data + errors
    """
    return data.__data_split(datax, datay, **kwargs)


def fit_split(datax, datay, **kwargs):
    """
    Splits datax and datay into (x,y,xerr,yerr).

    Parameters
    ----------
    **kwargs : optional
        see :func:`fit_kwargs`.
    """
    kwargs = fit_kwargs(kwargs)
    return data.filtered_data_split(datax, datay, **kwargs)


if __name__ == "__main__":
    import doctest

    doctest.testmod()

import uncertainties as unc


def _fit_minuit_leastsquares(datax, datay, function, yerr, params=None, **kwargs):
    # everything in iminuit is done through the Minuit object, so we import it
    from iminuit import Minuit

    # we also need a cost function to fit and import the LeastSquares function
    from iminuit.cost import LeastSquares
    from iminuit.util import Matrix

    # TODO check/add params
    if params is None:
        params = []
    least_squares = LeastSquares(
        datax, datay, yerr if yerr is not None else 1, function
    )
    m = Minuit(least_squares, *params)
    m.migrad()
    m.hesse()
    # print(m.values)
    # print(m.covariance)
    # print("chi2 = ", m.fval)
    # print("ndof = ", len(datax) - m.nfit)

    # fix slice issue from iminuites rewritten __getitem__ by using super
    return unc.correlated_values(m.values, super(Matrix, m.covariance))


import warnings

from scipy import optimize
from scipy.odr import ODR, Model, RealData

# fittet ein dataset mit gegebenen x und y werten, eine funktion und ggf. anfangswerten und y-Fehler
# gibt die passenden parameter der funktion, sowie dessen unsicherheiten zurueck
#
# https://stackoverflow.com/questionsquestions/14581358/getting-standard-errors-on-fitted-parameters-using-the-optimize-leastsq-method-i#
# Updated on 4/6/2016
# User: https://stackoverflow.com/users/1476240/pedro-m-duarte


def _fit_curvefit(datax, datay, function, params=None, yerr=None, **kwargs):
    try:
        pfit, pcov = optimize.curve_fit(
            function,
            datax,
            datay,
            p0=params,
            sigma=yerr,
            epsfcn=util.get("epsfcn", kwargs, 0.0001),
            **kwargs,
            maxfev=util.get("maxfev", kwargs, 10000),
        )
    except RuntimeError as e:
        debug.msg(str(e))
        return params
    error = []
    for i in range(len(pfit)):
        try:
            error.append(np.absolute(pcov[i][i]) ** 0.5)
        except Exception as e:
            warnings.warn(str(e))
            error.append(0.00)
    # print(pcov)
    # restore cov: unc.covariance_matrix([*ff])
    return unc.correlated_values(pfit, pcov)


# https://stackoverflow.com/a/52592811


def _fit_odr(datax, datay, function, params=None, yerr=None, xerr=None):
    model = Model(lambda p, x: function(x, *p))
    realdata = RealData(datax, datay, sy=yerr, sx=xerr)
    odr = ODR(realdata, model, beta0=params)
    out = odr.run()
    # This was the old wrong way! Now use correct co. matrix through unc-package!
    # Note Issues on scipy odr and curve_fit, regarding different definitions/namings of standard deviation or error and covaraince matrix
    # https://github.com/scipy/scipy/issues/6842
    # https://github.com/scipy/scipy/pull/12207
    # https://stackoverflow.com/questions/62460399/comparison-of-curve-fit-and-scipy-odr-absolute-sigma
    return unc.correlated_values(out.beta, out.cov_beta)



"""Simplified plotting."""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import smplr
import sympy
import uncertainties
import uncertainties.unumpy as unp
from matplotlib import pylab
from matplotlib.pyplot import *

# local imports
from smpl import doc, interpolate, io, stat, util, wrap
from smpl import fit as ffit


def set_plot_style():
    # fig_size = (8, 6)
    # fig_legendsize = 14
    # fig_labelsize = 12 # ‘xx-small’, ‘x-small’, ‘small’, ‘medium’, ‘large’, ‘x-large’, ‘xx-large’.
    params = {
        "legend.fontsize": "x-large",
        "figure.figsize": (8, 6),
        "axes.labelsize": "x-large",
        "axes.titlesize": "x-large",
        "xtick.labelsize": "x-large",
        "ytick.labelsize": "x-large",
    }
    pylab.rcParams.update(params)
    matplotlib.rcParams.update(params)
    # matplotlib.rcParams.update({'font.size': fig_labelsize})
    # colors = dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS)


unv = unp.nominal_values
usd = unp.std_devs

default = {
    #    'params'        :[None      ,"Initial fit parameters",
    "title": [None, "Plot title"],
    "xlabel": [
        "",
        "X axis label",
    ],
    "ylabel": [
        "",
        "Y axis label",
    ],
    "label": [
        None,
        "Legend name of plotted ``data``",
    ],
    "fmt": [
        ".",
        "Format for plotting fit function (could be a line style, marker or a combination of both or 'step')",
    ],
    "where": [
        "mid",
        "Where to place the ticks. Possible values are 'pre', 'post', 'mid', 'edge'. Might be incompatible with uncertainties.",
    ],
    "units": [
        None,
        "Units of the fit parameters as strings. Displayed in the Legend",
    ],
    "save": [
        None,
        " File to save the plot",
    ],
    "lpos": [
        0,
        "Legend position",
    ],
    "tight": [
        True,
        "tight_layout",
    ],
    #          'frange'        :[None      ,"Limit the fit to given range. First integer is the lowest and second the highest index.",],
    "prange": [
        None,
        "Limit the plot of the fit to given range",
    ],
    "sigmas": [
        0,
        "Color the array of given ``sigma`` times uncertainty. Only works if the fit function is coded with ``unp``",
    ],
    "data_sigmas": [
        1,
        "Color the array of given ``sigma`` times uncertainty. Only works if the data has uncertainties",
    ],
    "init": [False, "Initialize a new plot"],
    "ss": [
        True,
        "save, add legends and grid to the plot",
    ],
    "also_data": [True, " also plot the data"],
    "also_fit": [
        True,
        "also plot the fit",
    ],
    "auto_fit": [
        False,
        "automatically fit",
    ],
    "logy": [
        False,
        "logarithmic x axis",
    ],
    "logx": [
        False,
        "logarithmic y axis",
    ],
    "function_color": [
        None,
        "Color of the function plot",
    ],
    "data_color": [
        None,
        "Color of the data plot",
    ],
    "fit_color": [
        None,
        "Color of the fit plot",
    ],
    "fit_fmt": [
        "-",
        "Format of the fit plot",
    ],
    "residue": [
        False,
        "Display difference between fit and data in a second plot",
    ],
    "residue_err": [
        True,
        "Differences between fit and data will have errorbars",
    ],
    "show": [
        False,
        "Call plt.show()",
    ],
    "size": [
        None,
        "Size of the plot as a tuple (x,y). Only has an effect if ``init`` is True",
    ],
    "number_format": [
        io.gf(4),
        "Format to display numbers.",
    ],
    # ,          'selector'      :[ None     ,"Function that takes ``x`` and ``y`` as parameters and returns an array mask in order to limit the data points for fitting. Alternatively a mask for selecting elements from datax and datay.",],
    # ,          'fixed_params'  :[ True     ,"Enable fixing parameters by choosing the same-named variables from ``kwargs``.",],
    # ,          'sortbyx'       :[ True     , "Enable sorting the x and y data so that x is sorted.",],
    "interpolate": [False, "Enable interpolation of the data."],
    "interpolate_fmt": [
        "-",
        "Either format string or linestyle tuple.",
    ],
    "interpolate_label": [
        None,
        "Label for the interpolation.",
    ],
    "extrapolate": [
        True,
        "Enable extrapolation of whole data if fit range is limited by ``frange`` or ``fselector``.",
    ],
    "extrapolate_min": [
        None,
        "Lower extrapolation bound",
    ],
    "extrapolate_max": [
        None,
        "Higher extrapolation bound",
    ],
    "extrapolate_fmt": [
        "--",
        "Format of the extrapolation line",
    ],
    "extrapolate_hatch": [
        r"||",
        "Extrapolation shape/hatch for filled area in case of ``sigmas``>0. See https://matplotlib.org/stable/gallery/shapes_and_collections/hatch_style_reference.html",
    ],
    "bbox_to_anchor": [
        None,
        "Position in a tuple (x,y),Shift position of the legend out of the main pane. ",
    ],
    "ncol": [
        None,
        "Columns in the legend if used with ``bbox_to_anchor``.",
    ],
    "steps": [
        1000,
        "resolution of the plotted function",
    ],
    "fitinline": [
        False,
        "No newlines for each fit parameter",
    ],
    "grid": [
        True,
        "Enable grid for the plot",
    ],
    "hist": [
        False,
        "Enable histogram plot",
    ],
    "stairs": [
        False,
        "Enable stair plot",
    ],
    "capsize": [5, "size of cap on error bar plot"],
    "axes": [None, "set current axis"],
    "linestyle": [None, "linestyle, only active if `fmt`=None"],
    "xspace": [
        np.linspace,
        "xspace gets called with xspace(xmin,xmax,steps) in :func:`function` to get the points of the function that will be drawn.",
    ],
    "alpha": [0.2, "alpha value for the fill_between plot"],
    "append_chi2": [False, "Append chi2 to legend"],
    "append_r2": [False, "Append r2 to legend"],
    "append_dof": [False, "Append degrees of freedom to legend"],
}


@doc.append_doc(ffit.fit_kwargs)
@doc.append_str("\t")
@doc.append_str(doc.array_table(default, init=False))
@doc.append_str(
    doc.array_table({"plot_kwargs        ": ["default", "description"]}, bottom=False)
)
@doc.append_str("\n\n")
def plot_kwargs(kwargs):
    """Set default :func:`plot_kwargs` if not set."""
    kwargs = ffit.fit_kwargs(kwargs)
    for k, v in default.items():
        if k not in kwargs:
            kwargs[k] = v[0]
    return kwargs


# TODO optimize to minimize number of calls to function
# @append_doc(default_kwargs)


def fit(func, *adata, **kwargs):
    """
    Fit and plot function to datax and datay.

    Parameters
    ----------
    datax : array_like
        X data either as ``unp.uarray`` or ``np.array`` or ``list``
    datay : array_like
        Y data either as ``unp.uarray`` or ``np.array`` or ``list``
    function : func
        Fit function with parameters: ``x``, ``params``
    **kwargs : optional
        see :func:`plot_kwargs`.
    Fit parameters can be fixed via ``kwargs`` eg. ``a=5``.

    Returns
    -------
    array_like
        Optimized fit parameters of ``function`` to ``datax`` and ``datay``.
        If ``datay`` is complex, both the real and imaginary part are returned.

    Examples
    --------

    .. plot::
        :include-source:

        >>> from smpl import functions as f
        >>> from smpl import plot
        >>> param = plot.fit([0,1,2],[0,1,2],f.line)
        >>> float(plot.unv(param).round()[0])
        1.0

    """
    if "xaxis" in kwargs and ("xlabel" not in kwargs or not kwargs["xlabel"]):
        # warnings.warn("xaxis is deprecated. Use xlabel instead.", DeprecationWarning, 2)
        kwargs["xlabel"] = kwargs["xaxis"]
        # TODO maybe pop
    if "yaxis" in kwargs and ("ylabel" not in kwargs or not kwargs["ylabel"]):
        # warnings.warn("yaxis is deprecated. Use ylabel instead.", DeprecationWarning, 2)
        kwargs["ylabel"] = kwargs["yaxis"]
        # TODO maybe pop

    function = func
    if "function" in kwargs:
        function = kwargs["function"]
        del kwargs["function"]
        adata = [func, *adata]
    # Fix parameter order if necessary
    elif isinstance(function, (list, tuple, np.ndarray)):
        adata = [adata[-1], function, *adata[:-1]]
        function = adata[0]
        adata = adata[1:]
    if util.true("bins", kwargs):
        # yvalue will be overwritten
        ndata = [*adata, *adata]
        for i, o in enumerate(adata):
            ndata[2 * i] = o
            ndata[2 * i + 1] = o * 0
        adata = ndata

    assert len(adata) % 2 == 0, "data must be pairs of x and y data"
    if len(adata) == 2:
        datax, datay = adata
    else:
        rs = []
        for i in range(0, len(adata), 2):
            datax, datay = adata[i], adata[i + 1]
            if util.true("bins", kwargs):
                rs.append(fit(function, datax, **kwargs))
            else:
                rs.append(fit(function, datax, datay, **kwargs))
        return rs

    kwargs = plot_kwargs(kwargs)

    if np.any(np.iscomplex(datay)):
        label = util.get("label", kwargs, "")
        kwargs["label"] = label + "(real)"
        r = fit(datax, datay.real, function=function, **kwargs)
        kwargs["label"] = label + "(imag)"
        i = fit(datax, datay.imag, function=function, **kwargs)
        return r, i
    if kwargs["auto_fit"]:
        best_f, best_ff, lambda_f = ffit.auto(datax, datay, function, **kwargs)
        if best_f is not None:
            del kwargs["auto_fit"]
            fit(datax, datay, best_f, **kwargs)
        return best_f, best_ff, lambda_f
    if kwargs["also_fit"] == False and kwargs["label"] is None and kwargs["lpos"] == 0:
        kwargs["lpos"] = -1
    return _fit_impl(datax, datay, function, **kwargs)


def _fit_impl(datax, datay, function, **kwargs):
    x = None
    y = None
    rfit = None
    ifit = None
    fig = None
    fig = init_plot(kwargs)
    ll = None
    if kwargs["also_data"]:
        ll = plt_data(datax, datay, **kwargs).get_color()
    if kwargs["interpolate"]:
        ifit, _, x, y = plt_interpolate(datax, datay, icolor=ll, **kwargs)
    if kwargs["also_fit"]:
        assert function is not None, "function must be given"
        rfit, kwargs["fit_color"], _, _ = plt_fit(datax, datay, function, **kwargs)
    if kwargs["ss"]:
        kwargs["oldshow"] = kwargs["show"]
        kwargs["show"] = kwargs["show"] and not kwargs["residue"]
        save_plot(**kwargs)
        kwargs["show"] = kwargs["oldshow"]
    if kwargs["residue"] and fig is not None:
        plt_residue(datax, datay, function, rfit, fig, **kwargs)
    if not kwargs["also_fit"] and kwargs["interpolate"]:
        return (ifit, x, y)
        # return ifit
    return rfit


# @append_doc(default_kwargs)


def data(*data, function=None, **kwargs):
    """
    Plot datay against datax via :func:`fit`

    Parameters
    ----------
    datax : array_like
        X data either as ``unp.uarray`` or ``np.array`` or ``list``
    datay : array_like
        Y data either as ``unp.uarray`` or ``np.array`` or ``list``
    function : func,optional
        Fit function with parameters: ``x``, ``params``
    **kwargs : optional
        see :func:`plot_kwargs`.
    Returns
    -------
    array_like
        Optimized fit parameters of ``function`` to ``datax`` and ``datay``
    """
    if "also_fit" not in kwargs:
        kwargs["also_fit"] = False
    kwargs = plot_kwargs(kwargs)
    return fit(function=function, *data, **kwargs)


# @append_doc(default_kwargs)


def auto(*adata, funcs=None, **kwargs):
    """
    Automatically loop over functions and fit the best one.

    Parameters
    ----------
    funcs : function array
        functions to consider as fit. Default all ``smpl.functions``.
    **kwargs : optional
        see :func:`plot_kwargs`.

    Returns
    -------
    The best fit function and it's parameters. Also a lambda function where the parameters are already applied.



    """
    if "auto_fit" not in kwargs:
        kwargs["auto_fit"] = True
    kwargs = plot_kwargs(kwargs)
    return fit(function=funcs, *adata, **kwargs)


def _function(
    func,
    xfit,
    fmt="-",
    label=None,
    function_color=None,
    sigmas=0.0,
    alpha=0.4,
    **kwargs,
):
    # kargs = {}
    # if util.has("fmt", kwargs):
    #    kargs["fmt"] = kwargs["fmt"]
    # if util.has("label", kwargs) and kwargs["label"] != "":
    #    kargs["label"] = kwargs["label"]
    # if util.has("function_color", kwargs) and kwargs["function_color"] != "":
    #    kargs["color"] = kwargs["function_color"]
    # if util.has("sigmas", kwargs) and kwargs["sigmas"] != "":
    #    kargs["sigmas"] = kwargs["sigmas"]
    # if util.has("alpha", kwargs) and kwargs["alpha"] != "":
    #    kargs["alpha"] = kwargs["alpha"]
    # __function(func, xfit,  **kargs)

    if label == "":
        label = None
    if function_color == "":
        function_color = None
    if sigmas == "":
        sigmas = 0.0
    if alpha == "":
        alpha = 0.4
    __function(
        func,
        xfit,
        fmt=fmt,
        label=label,
        color=function_color,
        sigmas=sigmas,
        alpha=alpha,
        **kwargs,
    )


def plt_plt(x, y, fmt, color, label, linestyle, **kwargs):
    if linestyle is None and fmt is not None:
        return plt.plot(x, y, fmt, label=label, color=color, **kwargs)
    if linestyle is not None and fmt is None:
        return plt.plot(x, y, label=label, color=color, linestyle=linestyle, **kwargs)
    if linestyle is None and fmt is None:
        return plt.plot(x, y, label=label, color=color, **kwargs)
    # should not reach here
    raise ValueError(
        "Either fmt or linestyle must be given, but not both. fmt=%s, linestyle=%s"
        % (fmt, linestyle)
    )


def __function(
    gfunc,
    xlinspace,
    fmt="-",
    label=None,
    color=None,
    hatch=None,
    sigmas=0.0,
    linestyle=None,
    alpha=0.4,
    **kwargs,
):
    # filter unused bad kwargs here to avoid passing them down
    # TODO it would be better to not pass them down in the first place
    for key in [
        "pre",
        "post",
        "xaxis",
        "yaxis",
        "xvar",
        "xmin",
        "xmax",
        "xlabel",
        "ylabel",
        "bins",
        "binunc",
        "bbox_to_anchor",
        "tight",
        "residue",
        "lpos",
        "interpolate",
        "params",
        "also_fit",
        "init",
        "frange",
        "epsfcn",
        "units",
        "fselector",
        "maxfev",
        "sortbyx",
        "xerror",
        "yerror",
        "fixed_params",
        "autotqdm",
        "fitter",
        "title",
        "where",
        "save",
        "prange",
        "ss",
        "also_data",
        "auto_fit",
        "data_sigmas",
        "logy",
        "logx",
        "data_color",
        "fit_color",
        "fit_fmt",
        "show",
        "size",
        "number_format",
        "selector",
        "fitinline",
        "grid",
        "hist",
        "stairs",
        "capsize",
        "axes",
        "xspace",
        "extrapolate",
        "extrapolate_min",
        "extrapolate_max",
        "extrapolate_fmt",
        "extrapolate_hatch",
        "function_color",
        "residue_err",
        "interpolate_fmt",
        "interpolate_label",
        "interpolate_lower_uncertainty",
        "ncol",
        "steps",
        "interpolator",
        "next_color",
        "append_chi2",
        "append_r2",
        "append_dof",
    ]:
        kwargs.pop(key, None)
    func = gfunc
    x = xlinspace
    l = label

    if isinstance(func(x)[0], uncertainties.UFloat):
        if sigmas > 0:
            (ll,) = plt_plt(
                x,
                unv(func(x)),
                fmt,
                label=None,
                color=color,
                linestyle=linestyle,
                **kwargs,
            )
            y = func(x)
            plt.fill_between(
                x,
                unv(y) - sigmas * usd(y),
                unv(y) + sigmas * usd(y),
                alpha=alpha,
                label=l,
                color=ll.get_color(),
                hatch=hatch,
                **kwargs,
            )
        else:
            (ll,) = plt_plt(
                x,
                unv(func(x)),
                fmt,
                label=l,
                color=color,
                linestyle=linestyle,
                **kwargs,
            )
    else:
        (ll,) = plt_plt(
            x, func(x), fmt, label=l, color=color, linestyle=linestyle, **kwargs
        )
    return ll


def function(func, *args, fmt="-", **kwargs):
    """
    Plot function ``func`` between ``xmin`` and ``xmax``

    Parameters
    ----------
    func : function
        Function to be plotted between ``xmin`` and ``xmax``, only taking `array_like` ``x`` as parameter
    *args : optional
        arguments for ``func``
    **kwargs : optional
        see :func:`plot_kwargs`.
    """
    kwargs["fmt"] = fmt
    if not util.has("xmin", kwargs) or not util.has("xmax", kwargs):
        kwargs["xmin"], kwargs["xmax"] = stat.get_interesting_domain(func)
        # raise Exception("xmin or xmax missing.")

    # if not util.has('lpos', kwargs) and not util.has('label', kwargs):
    #    kwargs['lpos'] = -1

    if "label" not in kwargs:
        kwargs = plot_kwargs(kwargs)
        kwargs["label"] = get_fnc_legend(func, args, **kwargs)
    else:
        kwargs = plot_kwargs(kwargs)

    xlin = kwargs["xspace"](kwargs["xmin"], kwargs["xmax"], kwargs["steps"])
    init_plot(kwargs)

    # kwargs['lpos'] = 0
    # _plot(xfit, func(xfit, *args), **kwargs)
    _function(wrap.get_lambda_argd(func, kwargs["xvar"], *args), xlin, **kwargs)
    if kwargs["ss"]:
        save_plot(**kwargs)


# xaxis="",yaxis="",fit_color=None,save = None,residue_err=True,show=False):
def plt_residue(datax, datay, gfunction, rfit, fig, **kwargs):
    function = wrap.get_lambda(gfunction, kwargs["xvar"])
    fig.add_axes((0.1, 0.1, 0.8, 0.2))
    kwargs["ylabel"] = "$\\Delta$" + kwargs["ylabel"]
    kwargs["data_color"] = kwargs["fit_color"]

    if kwargs["residue_err"]:
        plt_data(datax, datay - function(datax, *rfit), **kwargs)
    else:
        plt_data(unv(datax), unv(datay - function(datax, *rfit)), **kwargs)
    kwargs["lpos"] = -1
    save_plot(**kwargs)


def data_split(datax, datay, **kwargs):
    return ffit.data_split(datax, datay, **kwargs)


def _fit(datax, datay, function, **kwargs):
    """
    Returns a fit like :func:`fit` but does no plotting.
    """
    return ffit.fit(datax, datay, function, **kwargs)


def plt_data(datax, datay, **kwargs):
    """
    Plot datay vs datax
    """
    x, y, xerr, yerr = data_split(datax, datay, **kwargs)
    if xerr is not None:
        xerr = xerr * kwargs["data_sigmas"]
    if yerr is not None:
        yerr = yerr * kwargs["data_sigmas"]

    ll = None
    if xerr is None and yerr is None:
        if kwargs["fmt"] is None:
            if kwargs["linestyle"] is None:
                (ll,) = plt.plot(
                    x, y, label=kwargs["label"], color=kwargs["data_color"]
                )
            else:
                (ll,) = plt.plot(
                    x,
                    y,
                    label=kwargs["label"],
                    color=kwargs["data_color"],
                    linestyle=kwargs["linestyle"],
                )
        elif kwargs["fmt"] == "step":
            (ll,) = plt.step(
                x,
                y,
                where=kwargs["where"],
                label=kwargs["label"],
                color=kwargs["data_color"],
            )
        elif kwargs["fmt"] == "hist":
            (ll,) = plt.step(
                x,
                y,
                where=kwargs["where"],
                label=kwargs["label"],
                color=kwargs["data_color"],
            )
            plt.fill_between(x, y, step="mid", color=ll.get_color())
        else:
            (ll,) = plt.plot(
                x, y, kwargs["fmt"], label=kwargs["label"], color=kwargs["data_color"]
            )
    elif kwargs["fmt"] is None:
        if kwargs["linestyle"] is None:
            (
                ll,
                _,
                _,
            ) = plt.errorbar(
                x,
                y,
                yerr=yerr,
                xerr=xerr,
                fmt=" ",
                capsize=kwargs["capsize"],
                label=kwargs["label"],
                color=kwargs["data_color"],
            )
        else:
            (
                ll,
                _,
                _,
            ) = plt.errorbar(
                x,
                y,
                yerr=yerr,
                xerr=xerr,
                fmt=" ",
                capsize=kwargs["capsize"],
                label=kwargs["label"],
                color=kwargs["data_color"],
                linestyle=kwargs["linestyle"],
            )
    elif kwargs["fmt"] == "step":
        (ll,) = plt.step(x, y, where=kwargs["where"], color=kwargs["data_color"])
        if xerr is not None:
            for ix, xv in enumerate(x):
                dx = xerr[ix]
                tx = [xv - dx, xv + dx]
                plt.fill_between(
                    tx,
                    y[ix] - yerr[ix],
                    y[ix] + yerr[ix],
                    label=kwargs["label"] if ix == 1 else None,
                    alpha=kwargs["alpha"],
                    step="pre",
                    color=ll.get_color(),
                )
        else:
            plt.fill_between(
                x,
                y - yerr,
                y + yerr,
                label=kwargs["label"],
                alpha=kwargs["alpha"],
                step="mid",
                color=ll.get_color(),
            )
    elif kwargs["fmt"] == "hist":
        (
            ll,
            _,
            _,
        ) = plt.errorbar(
            x,
            y,
            yerr=yerr,
            xerr=xerr,
            fmt=" ",
            capsize=kwargs["capsize"],
            color="black",
        )
        plt.fill_between(x, y, step="mid", label=kwargs["label"], color=ll.get_color())
    else:
        (
            ll,
            _,
            _,
        ) = plt.errorbar(
            x,
            y,
            yerr=yerr,
            xerr=xerr,
            fmt=kwargs["fmt"],
            capsize=kwargs["capsize"],
            label=kwargs["label"],
            color=kwargs["data_color"],
        )
    return ll


def get_fnc_legend(function, rfit, datax=None, datay=None, **kwargs):
    l = wrap.get_latex(function)

    vnames = wrap.get_varnames(function, kwargs["xvar"])
    for i in range(1, len(vnames)):
        l = l + ("\n" if not kwargs["fitinline"] or i == 1 else " ")
        l = l + "$" + sympy.latex(sympy.symbols(str(vnames[i]))) + "$="
        if kwargs["units"] is not None and usd(rfit[i - 1]) > 0:
            l = l + "("
        if "number_format" in kwargs:
            l = l + kwargs["number_format"].format(rfit[i - 1])
        else:
            l = l + "%s" % (rfit[i - 1])

        if kwargs["units"] is not None and usd(rfit[i - 1]) > 0:
            l = l + ")"
        if kwargs["units"] is not None:
            l = l + " " + kwargs["units"][i - 1]

    # Append Chi2, R2, and DOF if requested and data is available
    if datax is not None and datay is not None:
        if kwargs.get("append_chi2", False):
            try:
                chi2_val = ffit.Chi2(datax, datay, function, rfit, **kwargs)
                l = l + ("\n" if not kwargs["fitinline"] else " ")
                if "number_format" in kwargs:
                    l = l + "$\\chi^2$=" + kwargs["number_format"].format(chi2_val)
                else:
                    l = l + "$\\chi^2$=%s" % (chi2_val)
            except Exception:
                pass  # Ignore errors in Chi2 calculation

        if kwargs.get("append_r2", False):
            try:
                r2_val = ffit.R2(datax, datay, function, rfit, **kwargs)
                l = l + ("\n" if not kwargs["fitinline"] else " ")
                if "number_format" in kwargs:
                    l = l + "$R^2$=" + kwargs["number_format"].format(r2_val)
                else:
                    l = l + "$R^2$=%s" % (r2_val)
            except Exception:
                pass  # Ignore errors in R2 calculation

        if kwargs.get("append_dof", False):
            try:
                dof_val = ffit.Dof(datax, datay, function, rfit, **kwargs)
                l = l + ("\n" if not kwargs["fitinline"] else " ")
                if "number_format" in kwargs:
                    l = l + "DOF=" + kwargs["number_format"].format(dof_val)
                else:
                    l = l + "DOF=%s" % (dof_val)
            except Exception:
                pass  # Ignore errors in DOF calculation

    return l


def plt_fit_or_interpolate(
    datax, datay, fitted, l=None, c=None, f=None, ls=None, **kwargs
):
    # just filter these kwargs out, so they dont get passed down and are replaced by above args
    # TODO why not pass label=XXX directly to this?
    #      -> probably since there are cases where both e.g. color and replace color are needed
    for key in ["color", "label", "fmt", "linestyle", "hatch"]:
        kwargs.pop(key, None)
    if kwargs["prange"] is None:
        x, _, _, _ = ffit.fit_split(datax, datay, **kwargs)
        xfit = kwargs["xspace"](np.min(unv(x)), np.max(unv(x)), kwargs["steps"])
    else:
        xfit = kwargs["xspace"](
            kwargs["prange"][0], kwargs["prange"][1], kwargs["steps"]
        )
    ll = __function(
        fitted,
        xfit,
        kwargs["fit_fmt"] if f is not None and ls is None else f,
        label=l,
        color=kwargs["fit_color"] if c is None else c,
        linestyle=ls,
        **kwargs,
    )

    if (
        (
            (kwargs["frange"] is not None or kwargs["fselector"] is not None)
            and util.true("extrapolate", kwargs)
        )
        or util.has("extrapolate_max", kwargs)
        or util.has("extrapolate_min", kwargs)
    ):
        xxfit = kwargs["xspace"](
            util.get("extrapolate_min", kwargs, np.min(unv(datax))),
            util.get("extrapolate_max", kwargs, np.max(unv(datax))),
            kwargs["steps"],
        )
        for pmin, pmax in [
            (np.min(xxfit), np.min(xfit)),
            (np.max(xfit), np.max(xxfit)),
        ]:
            __function(
                fitted,
                kwargs["xspace"](pmin, pmax, kwargs["steps"]),
                util.get("extrapolate_fmt", kwargs, "--"),
                color=ll.get_color(),
                hatch=util.get("extrapolate_hatch", kwargs, r"||"),
                **kwargs,
            )
    return ll.get_color(), xfit, fitted(xfit)


def plt_interpolate(datax, datay, icolor=None, **kwargs):
    """
    Interpolate and Plot that Interpolation.
    """
    inter = interpolate.interpolate(datax, datay, **kwargs)
    kargs = {}
    if isinstance(kwargs["interpolate_fmt"], tuple):
        kargs["ls"] = kwargs["interpolate_fmt"]
    else:
        kargs["f"] = kwargs["interpolate_fmt"]
    if kwargs["interpolate_label"] is not None:
        kargs["l"] = kwargs["interpolate_label"]
    # l = None so that no label
    return (
        inter,
        *plt_fit_or_interpolate(datax, datay, inter, c=icolor, **kargs, **kwargs),
    )


def plt_fit(datax, datay, gfunction, **kwargs):
    """
    Fit and Plot that Fit.
    """
    func = wrap.get_lambda(gfunction, kwargs["xvar"])
    rfit = _fit(datax, datay, gfunction, **kwargs)

    def fitted(x):
        return func(x, *rfit)

    vnames = wrap.get_varnames(gfunction, kwargs["xvar"])

    l = get_fnc_legend(gfunction, rfit, datax, datay, **kwargs)
    for v in vnames[1:]:  # remove fixed parameters from kwargs # APN TODO why?
        kwargs.pop(v, None)

    return (rfit, *plt_fit_or_interpolate(datax, datay, fitted, l, **kwargs))


def init_plot(kwargs):
    fig = None
    if util.has("axes", kwargs) and kwargs["axes"] is not None:
        plt.sca(kwargs["axes"])
        fig = kwargs["axes"].get_figure()
    if kwargs["init"] or util.true("residue", kwargs):
        if kwargs["size"] is None:
            fig = plt.figure()
        else:
            fig = plt.figure(figsize=kwargs["size"])
        if kwargs["residue"]:
            fig.add_axes((0.1, 0.3, 0.8, 0.6))
    if util.has("next_color", kwargs) and not kwargs["next_color"]:
        lines = plt.gca()._get_lines
        tmp_color = lines._cycler_items[lines._idx]["color"]

        if kwargs["data_color"] is None:
            kwargs["data_color"] = tmp_color
        if kwargs["fit_color"] is None:
            kwargs["fit_color"] = tmp_color
        if kwargs["function_color"] is None:
            kwargs["function_color"] = tmp_color
    return fig


def save_plot(**kwargs):
    """
    save plot
    """
    smplr.style_plot1d(**kwargs)
    if "lpos" in kwargs and kwargs["lpos"] >= 0:
        if util.has("bbox_to_anchor", kwargs):
            if util.has("ncol", kwargs):
                plt.legend(
                    loc=kwargs["lpos"],
                    bbox_to_anchor=kwargs["bbox_to_anchor"],
                    ncol=kwargs["ncol"],
                    borderaxespad=0,
                )
            else:
                plt.legend(loc=kwargs["lpos"], bbox_to_anchor=kwargs["bbox_to_anchor"])
        else:
            plt.legend(loc=kwargs["lpos"])
    if "save" in kwargs and kwargs["save"] is not None:
        io.mkdirs(kwargs["save"])
        plt.savefig(kwargs["save"] + ".pdf")
    if kwargs.get("show"):
        show(**kwargs)


def show(**kwargs):
    kwargs = plot_kwargs(kwargs)

    smplr.style_plot1d(**kwargs)
    plt.show()


from smpl.plot2d import plot2d as _plot2d
from smpl.plot2d import plot2d_kwargs as _plot2d_kwargs


def plot2d(*args, **kwargs):
    _plot2d(*args, **kwargs)


def plot2d_kwargs(*args, **kwargs):
    _plot2d_kwargs(*args, **kwargs)



import pytest


def f(i):
    """
    >>> f(0)
    0
    """
    return i * 10


def g(n=10):
    """
    >>> g(10)
    450

    """
    return sum(f(i) for i in range(n))


@pytest.mark.line_profile.with_args(f, g)
def test_as_mark():

    assert g() == 450



from io import StringIO

from smpl.io.read_buffer import ReadBuffer


def cat(*inps, open=True):
    """
    Read all ``inps`` and return them as a single string.

    Parameters
    ----------
    inps : str, array_like, buffer
        object to read from

    Returns
    -------
    StringIO
        all ``inps`` as a StringIO buffer.

    Examples
    --------
    >>> from smpl import io
    >>> io.write("test.txt","hi\\nho1\\n2\\n3\\n4\\n")
    >>> io.cat("test.txt").getvalue()
    'hi\\nho1\\n2\\n3\\n4\\n'
    """
    r = StringIO()
    for inp in inps:
        with ReadBuffer(inp, open=open) as f:
            r.write(f.read())
    r.seek(0, 0)
    return r



class ReadBuffer:
    """
    Buffer for reading from a file if ``open`` is ``True``.
    """

    def __init__(self, input, open=False, **kwargs):
        self.input = input
        self.open = open
        self.file = None

    def __enter__(self):
        if self.open and isinstance(self.input, str):
            self.file = open(self.input)
            return self.file
        return self.input

    def __exit__(self, *args):
        if self.open and isinstance(self.input, str):
            self.file.close()



#!/usr/bin/env python3
"""Test script for append_dof functionality."""

import numpy as np
import matplotlib.pyplot as plt
from smpl import plot, functions

# Generate test data
x = np.linspace(0, 5, 20)
y = 2 * x + 1 + np.random.normal(0, 0.5, len(x))  # Linear with noise

print("Testing append_dof functionality...")

# Test 1: Basic fit without DOF
print("Test 1: Basic fit without DOF")
params1 = plot.fit(x, y, functions.line, label="Without DOF", init=True)
plt.title("Without DOF")
plt.show()

# Test 2: Fit with DOF appended
print("Test 2: Fit with DOF appended")
params2 = plot.fit(x, y, functions.line, label="With DOF", append_dof=True, init=True)
plt.title("With DOF")
plt.show()

# Test 3: Fit with Chi2, R2, and DOF all appended
print("Test 3: Fit with Chi2, R2, and DOF all appended")
params3 = plot.fit(x, y, functions.line, label="Complete stats", 
                   append_chi2=True, append_r2=True, append_dof=True, init=True)
plt.title("With Chi2, R2, and DOF")
plt.show()

print("DOF test completed!")



import matplotlib.pyplot as plt
import numpy as np
import pytest

from smpl import functions as f
from smpl import plot
from smpl.plot import init_plot


@pytest.mark.line_profile.with_args(plot.fit, plot.function, init_plot)
def test_fit():
    plot.function(
        lambda n, a, b: -a * n * np.log(b * n),
        1,
        1,
        xaxis="$N$",
        yaxis="$\\dot N$",
        xmin=0,
        xmax=100,
    )
    plot.show(block=False)
    plot.close()
    # print(io.pwd())

    data = np.loadtxt("tests/test_linear.txt")
    plot.fit(
        data[:, 0],
        data[:, 1],
        fmt=".",
        function=f.linear,
        units=["l", "b"],
        sigmas=1,
        lpos=2,
        residue=True,
        residue_err=False,
        xaxis="t",
        yaxis="s",
    )
    plot.fit(
        data[:, 0],
        data[:, 1],
        fmt=".",
        function=f.line,
        units=["l", "b"],
        sigmas=1,
        lpos=2,
        residue=True,
        residue_err=False,
        xaxis="t",
        yaxis="s",
        b=0,
    )

    plt.show(block=False)
    plot.close()



import matplotlib.pyplot as plt
import numpy as np
import pytest
from uncertainties import unumpy as unp

from smpl import plot


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_fit_str():
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    plot.fit("b*sin(x*c)", x, y)
    return plt.gcf()


@pytest.mark.mpl_image_compare(baseline_dir="baseline", remove_text=True)
def test_fit_lambda():
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    plot.fit(lambda a, b, c: b * unp.sin(a * c), x, y)
    return plt.gcf()



"""A list of various functions."""

import numpy as np
import scipy
import uncertainties.unumpy as unp

from smpl import doc


@doc.append_plot()
@doc.insert_str("Fac(x)")
def fac(n):
    return scipy.special.gamma(n + 1)


# general fit functions
@doc.append_plot(4)
@doc.insert_str("const(x) = $m$")
def const(x, m):
    # To get the right dimension
    return x - x + m


@doc.append_plot(2)
@doc.insert_latex_eq()
def linear(x, m):  # lineare Funktion mit f(x) = m * x
    return m * x


@doc.append_plot(2, -1)
@doc.insert_latex_eq()
def line(x, a, b):  # gerade mit = f(x) = m * x + b
    return a * x + b


# Gerade=line
# Line=line


@doc.append_plot(3, 0.02, 3)
@doc.insert_latex_eq()
def cos_abs(x, a, f, phi):
    return a * np.abs(unp.cos(2 * np.pi * f * (x - phi)))


@doc.append_plot(3, 0.02, 3)
@doc.insert_latex_eq()
def cos(x, a, f, phi):
    return a * unp.cos(2 * np.pi * f * (x - phi))


@doc.append_plot(3, 0.02, 3)
@doc.insert_latex_eq()
def sin(x, a, f, phi):
    return a * unp.sin(2 * np.pi * f * (x - phi))


@doc.append_plot(3, 0.02, 3)
@doc.insert_latex_eq()
def tan(x, a, f, phi):
    return a * unp.tan(2 * np.pi * f * (x - phi))


@doc.append_plot(0, 5, 3, 0)
@doc.insert_latex_eq()
def lorentz(x, x_0, a, d, y):
    return a / (np.pi * d * (1 + (x - x_0) ** 2 / d**2)) + y


# Lorentz=lorentz


@doc.append_plot(0, 5, 3, 0)
@doc.insert_latex_eq()
def gauss(x, x_0, a, d, y):
    return a * unp.exp(-((x - x_0) ** 2) / 2 / d**2) + y


Gauss = gauss


@doc.append_plot(0.5, 4)
@doc.insert_latex_eq()
def exponential(x, c, y_0):
    return unp.exp(c * x) * y_0


exp = exponential


@doc.append_plot(0.5, 4, xmin=0.1)
@doc.insert_latex_eq()
def log(x, c, y_0):
    return unp.log(c * x) * y_0


@doc.append_plot(1, 5, 0)
@doc.insert_latex_eq()
def square(x, x_0, a, y):
    return a * (x - x_0) ** 2 + y


@doc.append_plot(1, 5, 0, -2)
@doc.insert_latex_eq()
def cube(x, a, b, c, d):
    return a * x**3 + b * x**2 + c * x + d


@doc.append_plot(1, 3.3, 0)
@doc.insert_latex_eq()
def order(x, a, k, y):
    return a * (x) ** k + y


@doc.append_plot(1, 3.3, 0, xmin=0)
@doc.insert_latex_eq()
def sqrt(x, a, b, c):
    return a * unp.sqrt(x + b) + c



import matplotlib.pyplot as plt
import numpy as np
import pytest
import scipy
import tqdm

from smpl import animation, plot, stat
from smpl.plot import init_plot


def fac(n):
    return scipy.special.gamma(n + 1)


# from smpl.animation.animation import FigAnimation


# @pytest.mark.line_profile.with_args(plot.function,plot.fit,init_plot,FigAnimation.__init__)
@pytest.mark.line_profile.with_args(plot.function, plot.fit, init_plot)
def test_animation():
    plt.ioff()

    def update(a):
        plot.function(lambda x: a * x**2, xmin=0, xmax=5, init=True, tight=False)

    ani = animation.animate(
        update=update, frames=np.linspace(0, 10, 200), interval=10, blit=True
    )
    plt.show(block=False)
    plt.pause(10)
    plt.close("all")
    plt.ioff()
    for a in tqdm.tqdm(np.linspace(0, 10, 200)):
        plot.function(lambda x: a * x**2, xmin=0, xmax=5, init=True, tight=False)
        animation.frame()

    ani = animation.animate(interval=10, blit=True)
    plt.show(block=False)
    plt.pause(10)
    plt.close("all")
    ani.save("test.gif")


c = 0
datax = 0
datay = 0
bahnh = 0
bahnhs = 0


# @pytest.mark.line_profile.with_args(plot.function,plot.fit,init_plot,FigAnimation.__init__)
@pytest.mark.line_profile.with_args(plot.function, plot.fit, init_plot)
def test_histogram():
    global c, datax, datay, bahnh, bahnhs
    n = 13
    bahnhs = 13
    bahnh = np.array(range(bahnhs)) * 0
    datax = np.array(range(1, 14))
    datay = datax * 0
    c = 0

    def update(a):
        global c, bahnh
        c += 1
        bahnh = bahnh * 0
        for i in range(n):
            bahnh[np.random.randint(0, bahnhs)] += 1
            if np.any(np.greater(bahnh, 1)):
                datay[i] += 1
        plot.data(
            datax,
            stat.poisson_dist(datay) / c,
            init=True,
            tight=False,
            fmt="hist",
            ylabel="P(>1 Gäste bei beliebigem Bahnhof)",
            xlabel="$n$ Gäste",
            capsize=0,
        )
        plot.function(
            lambda ni: 1.0 - fac(bahnhs) / (fac(bahnhs - ni) * bahnhs**ni),
            xmin=0,
            xmax=13,
            label="$\\frac{b!}{(b-n)!b^n}$",
            function_color="orange",
        )

    animation.animate(
        update=update, frames=np.linspace(0, 20, 2000), interval=10, blit=False
    )
    plt.show(block=False)
    plt.pause(10)
    plt.close("all")


# @pytest.mark.line_profile.with_args(plot.function,plot.fit,init_plot,FigAnimation.__init__)
@pytest.mark.line_profile.with_args(plot.function, plot.fit, init_plot)
def test_subplots():
    plt.ioff()
    for a in tqdm.tqdm(np.linspace(0, 10, 200)):
        fig, axs = plot.subplots(1, 3, figsize=(12, 8), sharey=True)
        fig.subplots_adjust(wspace=0)
        plot.function(lambda x: a * x**2, axes=axs[0], tight=False, xmin=0, xmax=1)
        plot.function(lambda x: a * x**5, axes=axs[0], tight=False, xmin=0, xmax=1)
        plot.function(
            lambda x: a * x**3, axes=axs[1], logx=True, tight=False, xmin=0, xmax=1
        )
        plot.function(lambda x: a * x**1, axes=axs[2], tight=False, xmin=0, xmax=1)
        animation.frame()
    animation.animate(interval=10, blit=True)
    plt.show(block=False)
    plt.pause(10)
    plt.close("all")
    animation.clear()
    for a in tqdm.tqdm(np.linspace(0, 10, 100)):
        fig, axs = plot.subplots(1, 3, figsize=(12, 8), sharey=True)
        fig.subplots_adjust(wspace=0)
        plot.function(lambda x: a * x**3, axes=axs[1], logx=True, xmin=0, xmax=1)
        plot.function(lambda x: a * x**2, axes=axs[0], xmin=0, xmax=1)
        plot.function(lambda x: a * x**5, axes=axs[0], xmin=0, xmax=1)
        plot.function(lambda x: a * x**1, axes=axs[2], xmin=0, xmax=1)
        animation.frame()
    animation.animate(interval=10, blit=True)
    plt.show(block=False)
    plt.pause(10)
    plt.close("all")



"""Simplified python debuging."""

import os
from inspect import currentframe, getsource

import numpy as np

# TODO comapare against logging module https://docs.python.org/3.5/library/logging.html

# TODO instead of hard coded global variables, use function arguments


# TODO Enum maybe?
# DEBUG_LEVEL=-1 no debug
# DEBUG_LEVEL= 0 default level
# DEBUG_LEVEL= 1,2,3 higher levels, more debug
DEBUG_LEVEL = -1


# print prefix
DEBUG_PRE = "DBG"

# remove folders of files for print

# filename strings with .py (with folder if SPLIT_FILENAME==False)
WHITE_LIST_FILES = []
BLACK_LIST_FILES = []

active = debug = on = DEBUG_LEVEL >= 0
# count debug events by file+line
count_times = {}
cur_table_line = {}


def get_frame(_back=0):
    cf = currentframe()
    for _ in range(_back + 1):
        cf = cf.f_back
    return cf


def once(_back=0):
    """
    Returns true only one time

    Examples
    --------
    >>> for i in range(10):
    ...     if once():
    ...         print(i)
    0

    """
    return times(1, _back=_back + 1)


def times(t=1, _back=0):
    """
    Returns true if the count of the current line is greater than or equal to ``t``.

    Parameters
    ----------
    t : int
        The count to check against.
    _back : int
        Number of stack/frames to go back.

    Returns
    -------
    bool
        True if the count of the current line is greater than or equal to ``t``.

    Examples
    --------
    >>> reset_times()
    >>> for i in range(10):
    ...     if times(3):
    ...         print(i)
    0
    1
    2
    """
    line, fname = get_line_number_file(_back=_back + 1)
    inc_count(line, fname)
    return check_count(line, fname, t)


def get_line_src(_back=0):
    """
    Gets the current line in the python source.

    Parameters
    ----------
    _back : int
        Number of stack/frames to go back.

    Returns
    -------
    src : str
        The current line in the python source.

    Examples
    --------
    >>> get_line_src()
    'get_line_src()'
    >>> "funky"+get_line_src()
    'funky"funky"+get_line_src()'
    """
    cf = get_frame(_back=_back + 1)
    srcline = getsource(cf).split("\n")[cf.f_lineno - cf.f_code.co_firstlineno].strip()
    return srcline


def get_line_number_file(split=True, _back=0):
    """
    Gets the current filename and the current linenumber within it.

    Parameters
    ----------
    split : bool
        Indicates whenever the folders above of the file should be included in the returned filename.
    _back : int
        Number of stack/frames to go back.

    Returns
    -------
    filenumber : int
        First element in the return array
    filename : str
        Second element in the return array

    Examples
    --------
    >>> get_line_number_file()
    (1, '<doctest smpl.debug.debug.get_line_number_file[0]>')
    >>> for i in range(2):
    ...     get_line_number_file()
    (2, '<doctest smpl.debug.debug.get_line_number_file[1]>')
    (2, '<doctest smpl.debug.debug.get_line_number_file[1]>')
    """
    cf = get_frame(_back=_back + 1)
    fname = cf.f_code.co_filename
    if split:
        fname = cf.f_code.co_filename.split("/")[-1]
    return cf.f_lineno, fname


def get_line_number(_back=0):
    return get_line_number_file(_back=_back + 1)[0]


def line(msg_, tag="", level=0, times=-1, _back=0, **kwargs):
    msg(msg_, tag=tag, level=level, times=times, line_=True, _back=_back + 1, **kwargs)


# only once
def line1(msg_, tag="", level=0, times=1, _back=0, **kwargs):
    """
    Just like :func:`line` but ``times`` set to 1.

    Examples
    --------
    >>> for i in range(-2,2):
    ...     line1(i,level=-1)
    DBG::<doctest smpl.debug.debug.line1[0]>:2: line1(i,level=-1) = -2
    """
    msg1(msg_, tag=tag, level=level, times=times, line_=True, _back=_back + 1, **kwargs)


# counting functions


def get_count(line, fname):
    """
    Returns the counts of the line.

    Parameters
    ----------
    line : int
        The line in the python source of ``fname``.
    fname : str
        The filename.

    Returns
    -------
    count : int
        The count of the current line.

    Examples
    --------
    >>> get_count(1, "debug.py")
    0
    """
    global count_times
    if fname + ":" + str(line) in count_times:
        return count_times[fname + ":" + str(line)]
    return 0


def inc_count(line, fname):
    """
    Increments the count of the line.

    Parameters
    ----------
    line : int
        The line in the python source of ``fname``.
    fname : str
        The filename.

    Examples
    --------
    >>> inc_count(1, "debug.py")
    """
    global count_times
    if fname + ":" + str(line) in count_times:
        count_times[fname + ":" + str(line)] += 1
    else:
        count_times[fname + ":" + str(line)] = 1


def check_count(line, fname, t):
    """
    Returns true if the count of the line is greater than or equal to ``t``.

    Parameters
    ----------
    line : int
        The line in the python source of ``fname``.
    fname : str
        The filename.
    t : int
        The count to check against.

    Returns
    -------
    bool
        True if the count of the line is greater than or equal to ``t``.

    Examples
    --------
    >>> check_count(2, "debug.py", 0)
    True
    >>> inc_count(2, "debug.py")
    >>> check_count(2, "debug.py", 0)
    False

    """
    if t >= get_count(line, fname) or t == -1:
        if fname not in BLACK_LIST_FILES and (
            len(WHITE_LIST_FILES) == 0 or fname in WHITE_LIST_FILES
        ):
            return True
    return False


# _line enables printint src line
# t stands for times
def msg(msg, tag="", level=0, times=-1, line_=False, _back=0, **kwargs):
    """
    Prints the message ``msg`` if level > debug_level and always returns the msg.

    Parameters
    ----------
    tag : str
        Sets a tag to be printed for the debug message.
    level : int
        Debug level.
    times : int
        How often should the message be printed if the function gets called multiple times (e.g. in a loop).
    _line : bool
        Print the current line in the python source.
    _back : int
        Number of stack/frames to go back.

    Examples
    --------
    >>> msg("hi", level = -9999)
    DBG::<doctest smpl.debug.debug.msg[0]>:1: hi
    'hi'
    >>> msg("hi")
    'hi'

    """
    if level <= DEBUG_LEVEL:
        line, fname = get_line_number_file(_back=_back + 1)
        src = ""
        if line_:
            src = get_line_src(_back=_back + 1)
            src = src + " = "
        inc_count(line, fname)
        if check_count(line, fname, times):
            print(
                DEBUG_PRE
                + ":"
                + tag
                + ":"
                + fname
                + ":"
                + str(line)
                + ": "
                + src
                + str(msg)
            )
    return msg


def msg1(_msg, tag="", level=0, times=1, line_=False, _back=0, **kwargs):
    """
    Just like :func:`msg` but ``times`` set to 1.

    Parameters
    ----------
    tag : str
        Sets a tag to be printed for the debug message.
    level : int
        Debug level.
    times : int
        How often should the message be printed if the function gets called multiple times (e.g. in a loop).
    _line : bool
        Print the current line in the python source.
    _back : int
        Number of stack/frames to go back.

    Examples
    --------
    >>> for i in range(-2,2):
    ...     msg1(i, level = i)
    DBG::<doctest smpl.debug.debug.msg1[0]>:2: -2
    -2
    -1
    0
    1
    """
    return msg(
        _msg, level=level, tag=tag, times=times, line_=line_, _back=_back + 1, **kwargs
    )


def file(
    key,
    value,
    level=0,
    times=-1,
    seperator=";",
    _print=True,
    _back=0,
    filename="debug.csv",
):
    """
    Prints the message ``msg`` if level > debug_level to file ``filename``.
    """
    if level <= DEBUG_LEVEL:
        line, fname = get_line_number_file(_back=_back + 1)
        inc_count(line, fname)
        if check_count(line, fname, times):
            f = open(filename, "a+")
            if _print:
                tag = "debug.file"
                print(
                    DEBUG_PRE
                    + ":"
                    + tag
                    + ":"
                    + fname
                    + ":"
                    + str(line)
                    + ": "
                    + key
                    + seperator
                    + value
                )
            f.write(key + seperator + value + "\n")
            f.close()
    return value


def file1(_key, _value, level=0, times=1, _back=0, **kwargs):
    """
    Just like :func:`file` but ``times`` set to 1.
    """
    return file(_key, _value, level=level, times=times, _back=_back + 1, **kwargs)


def table_flush_header(filename="debug_table.csv", seperator=";"):
    """
    Saves the current keys from :func:`table` to ``filename``.
    """
    global cur_table_line
    f = open(filename, "a+")
    for key in sorted(cur_table_line):
        f.write(key + seperator)
    f.write("\n")
    f.close()


def table_flush_line(filename="debug_table.csv", seperator=";"):
    """
    Saves the current values from :func:`table` to ``filename``
    """
    global cur_table_line
    f = open(filename, "a+")
    ok = False
    dim = 1
    for cur in iter(cur_table_line):
        ok = isinstance(cur_table_line[cur], np.ndarray)
        if ok:
            dim = len(cur_table_line[cur])
            break
    for i in range(dim):
        for key in sorted(cur_table_line):
            if isinstance(cur_table_line[key], np.ndarray):
                f.write("%.30e" % cur_table_line[key][i] + seperator)
            else:
                f.write("%.30e" % cur_table_line[key] + seperator)
        f.write("\n")
    f.close()


def table(
    key,
    value,
    level=0,
    times=-1,
    seperator=";",
    _print=False,
    _back=0,
    filename="debug_table.csv",
):
    """
    Saves ``key``:``value`` in ``filename``.

    Examples
    --------
    >>> for i in range(-2,2):
    ...     table("a", i,level=-1)
    ...     table("b", i**2,level=-1)
    ...     table("c", i**i,level=-1)
    ...     if once(): table_flush_header();
    ...     table_flush_line()
    -2
    4
    0.25
    -1
    1
    -1.0
    0
    0
    1
    1
    1
    1
    >>> from smpl import io
    >>> print(io.read("debug_table.csv").strip())
    a;b;c;
    -2.000000000000000000000000000000e+00;4.000000000000000000000000000000e+00;2.500000000000000000000000000000e-01;
    -1.000000000000000000000000000000e+00;1.000000000000000000000000000000e+00;-1.000000000000000000000000000000e+00;
    0.000000000000000000000000000000e+00;0.000000000000000000000000000000e+00;1.000000000000000000000000000000e+00;
    1.000000000000000000000000000000e+00;1.000000000000000000000000000000e+00;1.000000000000000000000000000000e+00;
    >>> import pandas as pd
    >>> pd.read_csv("debug_table.csv")
                                                  a;b;c;
    0  -2.000000000000000000000000000000e+00;4.000000...
    1  -1.000000000000000000000000000000e+00;1.000000...
    2  0.000000000000000000000000000000e+00;0.0000000...
    3  1.000000000000000000000000000000e+00;1.0000000...
    >>> reset_table()
    >>> io.write("debug_table.csv","")
    >>> for i in range(1,3):
    ...     table("a", np.array([i*k for k in range(5)]),level=-1)
    ...     table("b", np.array([i*i*k for k in range(5)]),level=-1)
    ...     if once(): table_flush_header();
    ...     table_flush_line()
    array([0, 1, 2, 3, 4])
    array([0, 1, 2, 3, 4])
    array([0, 2, 4, 6, 8])
    array([ 0,  4,  8, 12, 16])
    >>> print(io.read("debug_table.csv").strip())
    a;b;
    0.000000000000000000000000000000e+00;0.000000000000000000000000000000e+00;
    1.000000000000000000000000000000e+00;1.000000000000000000000000000000e+00;
    2.000000000000000000000000000000e+00;2.000000000000000000000000000000e+00;
    3.000000000000000000000000000000e+00;3.000000000000000000000000000000e+00;
    4.000000000000000000000000000000e+00;4.000000000000000000000000000000e+00;
    0.000000000000000000000000000000e+00;0.000000000000000000000000000000e+00;
    2.000000000000000000000000000000e+00;4.000000000000000000000000000000e+00;
    4.000000000000000000000000000000e+00;8.000000000000000000000000000000e+00;
    6.000000000000000000000000000000e+00;1.200000000000000000000000000000e+01;
    8.000000000000000000000000000000e+00;1.600000000000000000000000000000e+01;
    """
    global cur_table_line
    if level <= DEBUG_LEVEL:
        line, fname = get_line_number_file(_back=_back + 1)
        inc_count(line, fname)
        if check_count(line, fname, times):
            if isinstance(value, np.ndarray):
                cur_table_line[key] = value.copy()
            else:
                cur_table_line[key] = value
    return value


def reset_times():
    """
    Resets global `count_times`.
    """
    global count_times
    count_times = {}


def reset_table():
    """
    Resets global `cur_table_line`.
    """
    global cur_table_line
    cur_table_line = {}


reset_count = reset_times

if os.path.exists("debug.csv"):
    os.remove("debug.csv")
if os.path.exists("debug_table.csv"):
    os.remove("debug_table.csv")

if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)



import line_profiler
import pytest

from smpl import functions, wrap

profiler = line_profiler.LineProfiler()


@profiler
def lin_fun(x, a_0, phi):
    return x * a_0 + phi


@pytest.mark.line_profile.with_args(wrap.get_latex)
def test_wrap():
    print(wrap.get_latex(lin_fun))
    print(wrap.get_latex(functions.linear))
    print(functions.linear.__doc__)
    print(
        wrap.get_latex(
            lambda a, b, c, x: (a + b + c) * x,
        )
    )



import matplotlib.pyplot as plt
import numpy as np

from smpl import plot


def gompertz(n, a, b):
    return -a * n * np.log(b * n)


def solution_gompertz(t, a, b, c):
    return 1 / b * np.exp(np.exp(-t * a) * c)


# @pytest.mark.line_profile.with_args(plot.function)
def test_em():
    for a in [1, 2]:
        for b in [1, 2]:
            plot.function(
                gompertz,
                a,
                b,
                xaxis="$N$",
                yaxis="$\\dot N$",
                xmin=0,
                xmax=5,
                init=False,
            )
    plt.show(block=False)
    plt.close()

    for a in [1, 2]:
        for b in [1, 2]:
            for c in [1, 2]:
                plot.function(
                    solution_gompertz,
                    a,
                    b,
                    c,
                    xaxis="$N$",
                    yaxis="$\\dot N$",
                    xmin=0,
                    xmax=5,
                    init=False,
                )
    plt.show(block=False)
    plt.close()



"""Simplified python code documentation."""

import warnings

from deprecation import deprecated as _deprecated


def append(txt):
    """
    Append ``txt`` in the ``target`` function docstring.

    Examples
    --------
    >>> @append('hi')
    ... def ho(): pass
    >>> ho.__doc__
    'hi'
    >>> @append(ho)
    ... def hi(): pass
    >>> hi.__doc__
    'hi'
    """
    if isinstance(txt, str):
        return append_str(txt)
    if hasattr(txt, "__doc__"):
        return append_doc(txt)
    warnings.warn(
        f"Can't append docstring from type {type(txt)}",
        stacklevel=2,  # Points to the caller of this function
    )
    return None


def append_str(txt):
    """
    Append ``txt`` in the ``target`` function docstring.

    Parameters
    ----------
    txt : ``str``
        ``txt`` is inserted to the ``__doc__`` of the ``target``

    Examples
    --------
    >>> def ho():
    ...     '''Ho'''
    ...     print(ho.__doc__)
    >>> @append_doc(ho)
    ... def hi():
    ...     '''Hi'''
    ...     print(hi.__doc__)
    >>> hi()
    HiHo
    """

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ += txt
        return target

    return wrapper


def append_doc(original):
    """
    Append doc string of ``original`` to ``target`` object.

    Parameters
    ----------
    original : ``class`` or ``function``
        ``original.__doc__`` is appended to the ``__doc__`` of the ``target``

    Examples
    --------
    >>> def ho():
    ...     '''Ho'''
    ...     print(ho.__doc__)
    >>> @append_doc(ho)
    ... def hi():
    ...     '''Hi'''
    ...     print(hi.__doc__)
    >>> hi()
    HiHo

    """
    return append_str(original.__doc__ if original.__doc__ is not None else "")


def insert(txt):
    """
    Insert ``txt`` in the ``target`` function docstring.

    Examples
    --------
    >>> @insert('hi')
    ... def ho(): pass
    >>> ho.__doc__
    'hi'
    >>> @insert(ho)
    ... def hi(): pass
    >>> hi.__doc__
    'hi'
    """
    if isinstance(txt, str):
        return insert_str(txt)
    if hasattr(txt, "__doc__"):
        return insert_doc(txt)
    warnings.warn(
        f"Can't append docstring from type {type(txt)}",
        stacklevel=2,  # Points to the caller of this function
    )
    return None


def insert_str(txt):
    """
    Insert ``txt`` in the ``target`` function docstring.

    Parameters
    ----------
    txt : ``str``
        ``txt`` is inserted to the ``__doc__`` of the ``target``

    Examples
    --------
    >>> def ho():
    ...     '''Ho'''
    ...     print(ho.__doc__)
    >>> @insert_doc(ho)
    ... def hi():
    ...     '''Hi'''
    ...     print(hi.__doc__)
    >>> hi()
    HoHi
    """

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ = txt + target.__doc__
        return target

    return wrapper


def insert_doc(original):
    """
    Inserts the docstring from passed function ``original`` in the ``target``
    function docstring.

    Parameters
    ----------
    original : ``class`` or ``function``
        ``original.__doc__`` is inserted to the ``__doc__`` of the ``target``

    Examples
    --------
    >>> def ho():
    ...     '''Ho'''
    ...     print(ho.__doc__)
    >>> @insert_doc(ho)
    ... def hi():
    ...     '''Hi'''
    ...     print(hi.__doc__)
    >>> hi()
    HoHi
    """
    return insert_str(original.__doc__)


def deprecated(
    version=None,
    deprecated_in=None,
    removed_in=None,
    reason=None,
    details=None,
):
    """
    Decorator to mark a function as deprecated.

    Parameters
    ----------
    version : ``str``
        Version of the package when the function was deprecated.
    deprecated_in : ``str``
        Version of the package when the function was deprecated.
    removed_in : ``str``
        Version of the package when the function will be removed.
    reason : ``str``
        Reason for deprecation.
    details : ``str``
        Details about the deprecation.

    Examples
    --------
    >>> @deprecated('0.0.0',removed_in='0.2.0')
    ... def ho():
    ...     '''Ho'''
    ...     print(ho.__doc__)
    >>> ho()
    Ho
    <BLANKLINE>
    .. deprecated:: 0.0.0
       This will be removed in 0.2.0.

    """
    # merge details and reason
    if details is None:
        details = reason
    elif reason is not None:
        details = reason + " " + details

    # merge deprecated_in and version
    if version is None:
        version = deprecated_in

    # increment minor version
    # if removed_in is None:
    #    removed_in = ".".join(
    #        [version.split(".")[0]]
    # + [str(int(version.split(".")[1]) + 2)] + ["0"]
    #    )

    return _deprecated(
        deprecated_in=version,
        removed_in=removed_in,
        # current_version=_version("pyfeyn2"),
        details=details,
    )


tab_len = 20


def table_sep(tabs=1):
    return (
        "=" * (tab_len - 2)
        + "  "
        + "=" * (tab_len - 2)
        + "  "
        + "=" * (tab_len - 2)
        + "\n"
        + "\t".join(["" for i in range(tabs + 1)])
    )


@deprecated(
    version="1.0.3.1",
    removed_in="?.?.?",
    reason="Use :func:`smpl.doc.array_table` instead.",
)
def table(dic, top=True, bottom=True, init=True, tabs=1):
    """
    Add dict= {'key': [values...]} to a simple reST table.

    Parameters
    ----------
    dic : ``dict``
        Dictionary to be converted to a table.
    top : ``bool``
        If ``True`` a top line is added.
    bottom : ``bool``
        If ``True`` a bottom line is added.
    init : ``bool``
        If ``True`` a tab is added at the beginning of the line.
    tabs : ``int``
        Number of tabs to be added at the beginning of the line.

    Returns
    -------
    ``str``
        The table as a string.
    >>> table({'a': [1, 2, 3], 'b': [4, 5, 6]})
    '\\t==================  ==================  ==================\\n\
\\ta                   1                   2                   3\\n\
\\tb                   4                   5                   6\\n\
\\t==================  ==================  ==================\\n\\t\\n'
    """
    t = "\t" * tabs
    rs = ""
    if init:
        rs += t
    if top:
        rs += table_sep(tabs=tabs)
    for k, vs in dic.items():
        rs += str(k)
        p = str(k)
        for v in vs:
            sv = str(v)
            if "<function " in sv and "at 0x" in sv:
                sv = sv.split("<function ")[1].split("at 0x")[0]
            rs += "".join([" " for i in range(tab_len - len(p))]) + sv
            p = sv
        rs += "\n" + t
    if bottom:
        rs += table_sep(tabs=tabs) + "\n"
    else:
        rs += ""
    return rs


def dict_to_table(dic):
    rt = []
    for k, vs in dic.items():
        rt += [[k, *vs]]
    return rt


def trim_eol_spaces(s):
    """
    Trim spaces at the end of lines.

    This is only needed for the docstrings, because black does trim them.

    """
    return "\n".join([ll.rstrip() for ll in s.split("\n")])


def array_table(arr, top=True, bottom=True, init=True, tabs=1, header=True):
    """
    Produces a reST table from a numpy array or normal 2d array.

    Parameters
    ----------
    arr : ``numpy.ndarray``, ``list`` or ``dict``
        2d array or dict
    top : ``bool``
        If ``True`` a top line is added.
    bottom : ``bool``
        If ``True`` a bottom line is added.
    init : ``bool``
        If ``True`` a tab is added at the beginning of each line.
    tabs : ``int``
        Number of tabs at the beginning of each line.
    header : ``bool``
        If ``True`` the first row is used as header.

    Examples
    --------
    >>> print(trim_eol_spaces(array_table([["a","b"],["hihi", "hoho"]]\
        ,tabs=0)))
    ====== ======
    a      b
    ====== ======
    hihi   hoho
    ====== ======
    <BLANKLINE>
    >>> print(trim_eol_spaces(array_table({"a":["hihi",2],"b": ["hoho",3]}\
        ,tabs=0,header=False)))
    === ====== ===
    a   hihi   2
    b   hoho   3
    === ====== ===
    <BLANKLINE>

    """
    if isinstance(arr, dict):
        arr = dict_to_table(arr)
    width = len(arr[0])
    height = len(arr)
    widths = [0 for i in range(width)]
    # maximum width of each column
    for i in range(width):
        for j in range(height):
            widths[i] = max(len(str(arr[j][i])), widths[i])
    rs = ""
    if init:
        rs += "\t" * tabs
    if top:
        rs += " ".join(["=" * (widths[i] + 2) for i in range(width)]) + "\n"

    for i in range(height):
        rs += "\t" * tabs
        for j in range(width):
            rs += str(arr[i][j]) + " " * (widths[j] - len(str(arr[i][j])) + 3)
        rs += "\n"
        if header and i == 0:
            rs += (
                "\t" * tabs
                + " ".join(["=" * (widths[i] + 2) for i in range(width)])
                + "\n"
            )

    if bottom:
        rs += "\t" * tabs
        rs += " ".join(["=" * (widths[i] + 2) for i in range(width)]) + "\n"
    return rs


if __name__ == "__main__":
    import doctest

    doctest.testmod()
